# Type dependencies
import os
import re
import threading
from contextlib import contextmanager, ContextDecorator
from itertools import chain
from typing import Optional, Iterable
from abc import ABC, abstractmethod
from .checkpointing import CheckpointManager, CheckpointError, StorageType
from .ordered_set import OrderedSet

_working_tape = None
_annotation_enabled = False


def get_working_tape():
    return _working_tape


def pause_annotation():
    """Switch off annotation."""
    global _annotation_enabled
    _annotation_enabled = False


def continue_annotation():
    """Switch on annotation."""
    global _annotation_enabled
    _annotation_enabled = True
    return _annotation_enabled


class set_working_tape(ContextDecorator):
    """Set a new tape as the working tape.

    This class can be used in three ways:
       1) as a free function to replace the working tape,
       2) as a context manager within which a new tape is set as the working tape,
       3) as a function decorator so that the new tape is set only inside the function.

       Example usage:

        1) Set a new tape as the working tape:

            .. highlight:: python
            .. code-block:: python

                set_working_tape(Tape())

        2) Set a local tape within a context manager:

            .. highlight:: python
            .. code-block:: python

                with set_working_tape() as tape:
                    ...

        3) Set the local tape inside a decorated function.
           The two functions below are equivalent:

            .. highlight:: python
            .. code-block:: python

                @set_working_tape()
                def decorated_function(*args, **kwargs):
                    # do something here
                    return ReducedFunctional(functional, control)

                def context_function(*args, **kwargs):
                    with set_working_tape():
                        # do something here
                        return ReducedFunctional(functional, control)

    """

    def __init__(self, tape=None, **tape_kwargs):
        # Get working tape
        global _working_tape
        # Store current tape
        self.old_tape = _working_tape
        # Set new tape
        self.tape = tape or Tape(**tape_kwargs)
        _working_tape = self.tape

    def __enter__(self):
        return self.tape

    def __exit__(self, *args):
        # Re-establish the original tape
        global _working_tape
        _working_tape = self.old_tape


class stop_annotating(ContextDecorator):
    """A context manager and function decorator within which annotation is stopped.

    Args:
        modifies (OverloadedType or list[OverloadedType]): One or more
            variables which appear in the tape and whose values are to be
            changed inside the context manager.

    The `modifies` argument is intended to be used by user code which
    changes the value of inputs to the adjoint calculation such as time varying
    forcings. Its effect is to create a new block variable for each of the
    modified variables at the end of the context manager. """

    def __init__(self, modifies=None):
        self.modifies = modifies
        # the `no_annotations` context manager could be nested,
        # so we need a stack to keep track of the original states.
        self._orig_annotation_enabled = []

    def __enter__(self):
        global _annotation_enabled
        if self.modifies and len(self._orig_annotation_enabled) != 0:
            raise ValueError(
                "Cannot use `modifies` argument if `stop_annotating` is nested,"
                " e.g. if used as the `no_annotations` decorator.")
        self._orig_annotation_enabled.append(_annotation_enabled)
        _annotation_enabled = False

    def __exit__(self, *args):
        global _annotation_enabled
        _annotation_enabled = self._orig_annotation_enabled.pop()
        if self.modifies is not None:
            try:
                self.modifies.clear_block_variable()
            except AttributeError:
                for var in self.modifies:
                    var.clear_block_variable()


no_annotations = stop_annotating()
"""Decorator to turn off annotation for the decorated function."""


def annotate_tape(kwargs=None):
    """Return True if annotation flag is on, and False if not.

    If kwargs is given, the function will try to extract the
    annotate keyword. If the annotate keyword is not present it defaults to True.
    If annotation has been paused, then it will always return False.

    Args:
        kwargs (dict): A dictionary of keyword arguments to extract from.
            Note that this should be passed as a dictionary and not actual keyword arguments.

    Returns: bool

    """
    annotate = kwargs is None or kwargs.pop("annotate", True)

    # TODO: Consider if there is any scenario where one would want the keyword to have
    # precedence over the global flag.
    if not _annotation_enabled:
        return False

    return annotate


class Tape(object):
    """The tape.

    The tape consists of blocks, :class:`Block` instances.
    Each block represents one operation in the forward model.

    """
    __slots__ = ["_blocks", "_tf_tensors", "_tf_added_blocks", "_nodes",
                 "_tf_registered_blocks", "_bar", "_package_data",
                 "_checkpoint_manager", "latest_checkpoint",
                 "_eagerly_checkpoint_outputs", "_recompute_count"]

    def __init__(self, blocks=None, package_data=None):
        # Initialize the list of blocks on the tape.
        self._blocks = TimeStepSequence(blocks=blocks)
        # Dictionary of TensorFlow tensors. Key is id(block).
        self._tf_tensors = {}
        # Keep a list of blocks that has been added to the TensorFlow graph
        self._tf_added_blocks = []
        self._tf_registered_blocks = []
        self._bar = _NullProgressBar
        # Hook location for packages which need to store additional data on the
        # tape. Packages should store the data under a "packagename" key.
        self._package_data = package_data or {}
        # Default to checkpointing all block variables.
        self.latest_checkpoint = float("inf")
        self._checkpoint_manager = None
        # Whether to store the adjoint dependencies.
        self._eagerly_checkpoint_outputs = False
        # A counter for the number of tape recomputations.
        self._recompute_count = 0

    def clear_tape(self):
        """Clear the tape."""
        self.reset_variables()
        self._blocks = TimeStepSequence()
        for data in self._package_data.values():
            data.clear()
        self._checkpoint_manager = None
        self._recompute_count = 0

    @property
    def latest_timestep(self):
        """The current time step to which blocks will be added."""
        return max(len(self._blocks.steps) - 1, 0)

    @property
    def recompute_count(self):
        """The number of times the tape has been recomputed."""
        return self._recompute_count

    def end_timestep(self):
        """Mark the end of a timestep when taping the forward model."""
        if self._checkpoint_manager:
            self._checkpoint_manager.end_timestep(self.latest_timestep)
        else:
            self._blocks.append_step()

    def timestepper(self, iterable):
        """Return an iterator that advances the tape timestep.

        Note:
            This method facilitates taping timestepping simulations so that recompute
            checkpointing can be used on the tape. For example, a simulation with
            10 timesteps might use a timestepping loop of this form::

                tape = get_working_tape()

                for timestep in tape.timestepper(range(10)):
                    ...

            This has the effect of calling `tape.end_timestep()` after each iteration.

        Args:
            iterable (iterable): The iterable definining the sequence of timesteps.

        Returns:
            TapeTimeStepper: An iterator that advances the tape timestep.
        """
        return TapeTimeStepper(self, iterable)

    def reset_blocks(self):
        """Calls the Block.reset method of all blocks on the tape.

        Also resets any package-dependent data stored on the tape.
        """
        for block in self._blocks:
            block.reset()
        for data in self._package_data.values():
            data.reset()

    def add_block(self, block):
        """
        Adds a block to the tape and returns the index.
        """
        self._blocks.append(block)

        # len() is computed in constant time, so this should be fine.
        return len(self._blocks) - 1

    def add_to_checkpointable_state(self, block_var, last_used):
        """Add a block variable into the checkpointable state set.

        Note:
            `checkpointable_state` is a set of block variables which are needed
            to restart from the start of a timestep.

        Args:
            block_var (BlockVariable): The block variable to add.
            last_used (int): The last timestep in which the block variable was used.
        """
        if not self.timesteps:
            self._blocks.append_step()
        for step in self.timesteps[last_used + 1:]:
            step.checkpointable_state.add(block_var)

    def add_to_adjoint_dependencies(self, block_var, last_used):
        """Add a block variable into the adjoint dependencies set.

        Note:
            `adjoint_dependencies` is a set of block variables which are needed
            to compute the adjoint of a timestep.

        Args:
            block_var (BlockVariable): The block variable to add.
            last_used (int): The last timestep in which the block variable was used.
        """
        if not self.timesteps:
            self._blocks.append_step()
        for step in self.timesteps[last_used + 1:]:
            step.adjoint_dependencies.add(block_var)

    def enable_checkpointing(self, schedule, gc_timestep_frequency=None, gc_generation=2):
        """Enable checkpointing on the adjoint evaluation.

        A checkpoint manager able to execute the forward and adjoint computations
        according to the schedule provided by checkpoint_schedules package.

        Args:
            schedule (checkpoint_schedules.schedule): A schedule provided by the
            checkpoint_schedules package.
            gc_timestep_frequency (None or int): The timestep frequency for garbage collection.
            For additional information, please refer to the :class:`CheckpointManager`
            documentation.
            gc_generation (int): The generation for garbage collection. For additional
            information, please refer to the :class:`CheckpointManager` documentation.
        """
        if self._blocks:
            raise CheckpointError(
                "Checkpointing must be enabled before any blocks are added to the tape."
            )

        if gc_timestep_frequency is not None and not isinstance(gc_timestep_frequency, int):
            raise CheckpointError("gc_timestep_frequency must be an integer.")

        self._checkpoint_manager = CheckpointManager(
            schedule, self, gc_timestep_frequency=gc_timestep_frequency,
            gc_generation=gc_generation)

    def get_blocks(self, tag=None):
        """Returns a list of the blocks on the tape.

        Use the kwarg `tag` to optionally get all
        blocks with a specified tag.

        Returns:
            list[block.Block]: A list of :class:`Block` instances.

        """
        if tag is None:
            return self._blocks
        else:
            return [block for block in self._blocks if block.tag == tag]

    def get_tags(self):
        """
        Returns a list of the unique tags used in blocks on the tape.
        """
        tags = []
        for block in self._blocks:
            if block.tag is not None and block.tag not in tags:
                tags.append(block.tag)
        return tags

    def evaluate_adj(self, last_block=0, markings=False):
        """Evaluate the adjoint of the tape.

        Args:
            last_block (int, optional): The index of the last block to evaluate.
            markings (bool, optional): If True, then each `BlockVariable` of the current block
                will have set `is_control_dependent` attribute indicating whether their adjoint
                components are relevant for computing the final target adjoint values and
                `is_functional_dependency` attribute indicating whether its value impacts the
                value of the functional.
        """
        if self._checkpoint_manager:
            self._checkpoint_manager.evaluate_adj(last_block, markings)
        else:
            for i in self._bar("Evaluating adjoint").iter(
                range(len(self._blocks) - 1, last_block - 1, -1)
            ):
                self._blocks[i].evaluate_adj(markings=markings)

    def evaluate_tlm(self, markings=False):
        for i in self._bar("Evaluating TLM").iter(
            range(len(self._blocks))
        ):
            self._blocks[i].evaluate_tlm(markings=markings)

    def evaluate_hessian(self, markings=False):
        for i in self._bar("Evaluating Hessian").iter(
            range(len(self._blocks) - 1, -1, -1)
        ):
            self._blocks[i].evaluate_hessian(markings=markings)

    def reset_variables(self, types=None):
        for i in range(len(self._blocks) - 1, -1, -1):
            self._blocks[i].reset_variables(types)

    def reset_hessian_values(self):
        for i in range(len(self._blocks) - 1, -1, -1):
            self._blocks[i].reset_variables(types=("hessian"))

    def reset_tlm_values(self):
        for i in range(len(self._blocks) - 1, -1, -1):
            self._blocks[i].reset_variables(types=("tlm"))

    def copy(self):
        """Returns a shallow copy of the tape.

        Returns:
            Tape: The copy of the tape.

        """
        # TODO: Offer deepcopying. But is it feasible memory wise to copy all checkpoints?
        return Tape(
            blocks=self._blocks,
            package_data={k: v.copy() for k, v in self._package_data.items()}
        )

    def checkpoint_block_vars(self, controls=[], tag=None):
        """Returns an object to checkpoint the current state of all block variables on the tape.

        Args:
            controls (list): A list of controls for which the block variables should also be extracted.

        Returns:
            dict[BlockVariable, object]: The checkpointed block variables of the tape.

        """

        state_dict = {
            var: var.checkpoint
            for var in chain(
                chain.from_iterable(b.get_outputs() for b in self.get_blocks(tag)),
                (control.block_variable for control in controls))
        }
        state_dict["package_data"] = {k: v.checkpoint() for k, v in self._package_data.items()}
        return state_dict

    def restore_block_vars(self, block_vars):
        """Set the checkpoints of the tape according to a checkpoint dictionary.

        Args:
            block_vars (dict[BlockVariable, object]): A checkpoint object from checkpoint_block_vars.

        """
        block_vars = block_vars.copy()
        package_data = block_vars.pop("package_data")

        for k, v in block_vars.items():
            # we use the private _checkpoint attribute
            # here because the public attribute is a no-op
            # if the control values are "active", but we
            # need to make sure they are reset to the
            # cached value as well
            k._checkpoint = v

        for k, v in self._package_data.items():
            v.restore_from_checkpoint(package_data[k])

    def optimize(self, controls=None, functionals=None):
        if controls is not None:
            self.optimize_for_controls(controls)

        if functionals is not None:
            self.optimize_for_functionals(functionals)

    def optimize_for_controls(self, controls):
        # TODO: Consider if we want Enlist wherever it is possible. Like in this case.
        # TODO: Consider warning/message on empty tape.
        nodes = set([control.block_variable for control in controls])
        discarded_variables = set()
        optimized_timesteps = TimeStepSequence()

        for step in self._blocks.steps:
            optimized_timesteps.append_step()

            for block in step:
                depends_on_control = False
                for dep in block.get_dependencies():
                    if dep in nodes:
                        depends_on_control = True
                        break

                if depends_on_control:
                    for output in block.get_outputs():
                        if output in nodes:
                            raise RuntimeError("Control depends on another control.")
                        nodes.add(output)
                    optimized_timesteps.append(block)
                else:
                    discarded_variables.union(block.get_outputs())
            optimized_timesteps.steps[-1].checkpointable_state = \
                step.checkpointable_state - discarded_variables

        self._blocks = optimized_timesteps

    def optimize_for_functionals(self, functionals):
        retained_nodes = set([functional.block_variable
                             for functional in functionals]
                             )
        optimized_timesteps = []

        for step in reversed(self._blocks.steps):
            current_blocks = []
            for block in reversed(step):
                produces_functional = False
                for dep in block.get_outputs():
                    if dep in retained_nodes:
                        produces_functional = True

                if produces_functional:
                    for dep in block.get_dependencies():
                        retained_nodes.add(dep)
                    current_blocks.append(block)
            optimized_timesteps.append(TimeStep(reversed(current_blocks)))

        optimized_timesteps.reverse()

        for step, new_step in zip(self._blocks.steps, optimized_timesteps):
            new_step.checkpointable_state = \
                step.checkpointable_state & retained_nodes

        self._blocks = TimeStepSequence(steps=optimized_timesteps)

    @contextmanager
    def marked_control_dependents(self, controls):
        """Mark all the block variables on which the given controls depend."""
        nodes = self._nodes_dependent_on_controls(controls)
        for node in nodes:
            node.is_control_dependent = True
        try:
            yield
        finally:
            for node in nodes:
                node.is_control_dependent = False

    def _nodes_dependent_on_controls(self, controls):
        # This method is just a stripped down Block.optimize_for_controls
        nodes = set([control.block_variable for control in controls])

        for block in self.get_blocks():
            depends_on_control = False
            for dep in block.get_dependencies():
                if dep in nodes:
                    depends_on_control = True

            if depends_on_control:
                for output in block.get_outputs():
                    nodes.add(output)
        return nodes

    @contextmanager
    def marked_functional_dependencies(self, functional):
        """Mark all of the block variables on which the functional depends."""
        nodes = self._functional_dependencies(functional)
        for node in nodes:
            node.is_functional_dependency = True
        try:
            yield
        finally:
            for node in nodes:
                node.is_functional_dependency = False

    def _functional_dependencies(self, functional):
        # This function is just a stripped down Block.optimize_for_controls
        nodes = set([functional.block_variable])

        for block in reversed(self.get_blocks()):
            functional_dependency = False
            for output in block.get_outputs():
                if output in nodes:
                    functional_dependency = True

            if functional_dependency:
                for dep in block.get_dependencies():
                    nodes.add(dep)
        return nodes

    @property
    def timesteps(self):
        """Return the list of time steps on this tape."""
        return self._blocks.steps

    def _valid_tf_scope_name(self, name):
        """Return a valid TensorFlow scope name"""
        valid_name = ""
        p = re.compile("[A-Za-z0-9_.\\-]")
        for ch in name:
            match = p.match(ch)
            if not match:
                if valid_name and valid_name[-1] != "_":
                    valid_name += "_"
            else:
                valid_name += ch
        return valid_name

    def _get_tf_scope_name(self, block):
        """Return a TensorFlow scope name based on the block's class name."""
        # If the block is a BlockVariable we use the class name of block.output
        if block.__class__.__name__ == "BlockVariable":
            if block.output.__class__.__name__ in ("AdjFloat",):
                name = str(block.output.__class__.__name__) + "_" + str(block)
            else:
                name = str(block.output.__class__.__name__)
        else:
            name = block.__class__.__name__
        return self._valid_tf_scope_name(name)

    def _tf_register_blocks(self, name=None):
        lst = list()
        lst.append(name)
        for block in self.get_blocks():
            if block in self._tf_added_blocks:
                continue
            self._tf_added_blocks.append(block)
            lst.append(block)
        self._tf_registered_blocks.append(lst)

    def _tf_rebuild_registered_blocks(self):
        """Remove blocks that no longer exist on the tape from registered blocks."""
        new_registered_blocks = []
        new_added_blocks = []
        for scope in self._tf_registered_blocks:
            lst = [scope[0]]
            for i in range(1, len(scope)):
                block = scope[i]
                if block in self.get_blocks():
                    lst.append(block)
                    new_added_blocks.append(block)

            if len(lst) > 1:
                new_registered_blocks.append(lst)
        self._tf_registered_blocks = new_registered_blocks
        self._tf_added_blocks = new_added_blocks

    def _tf_add_blocks(self):
        """Add new blocks to the TensorFlow graph."""

        import tensorflow as tf

        self._tf_register_blocks()
        self._tf_rebuild_registered_blocks()

        for scope in self._tf_registered_blocks:
            scope_name = scope[0]
            with tf.name_scope(scope_name):
                for i in range(1, len(scope)):
                    block = scope[i]

                    # Block dependencies
                    in_tensors = []
                    for dep in block.get_dependencies():
                        if id(dep) in self._tf_tensors:
                            in_tensors.append(self._tf_tensors[id(dep)])
                        else:
                            with tf.name_scope(self._get_tf_scope_name(dep)):
                                tin = tf.py_func(lambda: None, [], [tf.float64],
                                                 name=self._valid_tf_scope_name(str(dep)))
                                in_tensors.append(tin)
                                self._tf_tensors[id(dep)] = tin

                    # Block node
                    with tf.name_scope(self._get_tf_scope_name(block)):
                        tensor = tf.py_func(lambda: None, in_tensors, [tf.float64],
                                            name=self._valid_tf_scope_name(str(block)))
                        self._tf_tensors[id(block)] = tensor

                    # Block outputs
                    for out in block.get_outputs():
                        with tf.name_scope(self._get_tf_scope_name(out)):
                            tout = tf.py_func(lambda: None, [tensor], [tf.float64],
                                              name=self._valid_tf_scope_name(str(out)))
                            self._tf_tensors[id(out)] = tout

    @contextmanager
    def name_scope(self, name=None):
        """Returns a context manager that creates hierarchical names for TensorFlow operations.

        Args:
            name (str|None): Name of scope to use. Default None.
        """
        self._tf_register_blocks()
        yield
        self._tf_register_blocks(name)

    def visualise(self, output="log", launch_tensorboard=False, open_in_browser=False):
        """Makes a visualisation of the tape as a graph using TensorFlow
        or GraphViz. (Default: Tensorflow). If `output` endswith `.dot` or
        `.pdf`, Graphviz is used.

        Args:
            output (str): Directory where event files for TensorBoard is
                stored. Default log.
            launch_tensorboard (bool): Launch TensorBoard in the background.
                Default False.
            open_in_browser (bool): Opens http://localhost:6006/ in a web
                browser. Default False.
        """
        if output.endswith(".dot"):
            return self.visualise_dot(output)
        elif output.endswith(".pdf"):
            return self.visualise_pdf(output)

        import tensorflow as tf
        tf.compat.v1.reset_default_graph()
        self._tf_add_blocks()

        # Write graph to file
        with tf.compat.v1.Session() as sess:
            writer = tf.compat.v1.summary.FileWriter(output, sess.graph)
            writer.close()

        if not launch_tensorboard or not open_in_browser:
            print("Run the command line:\n"
                  "--> tensorboard --logdir={}\n"
                  "Then open http://localhost:6006/ in your web browser.".format(output))

        if launch_tensorboard:
            def launchTensorBoard():
                os.system('tensorboard --logdir=' + output)

            t = threading.Thread(target=launchTensorBoard, args=([]))
            t.start()

        if open_in_browser:
            import webbrowser
            webbrowser.open_new_tab("http://localhost:6006/")

    def create_graph(self, backend="networkx"):
        import networkx as nx
        G = nx.DiGraph()
        for i, block in enumerate(self._blocks):
            block.create_graph(G, pos=i)
        return G

    def visualise_dot(self, filename):
        """Makes a visualisation of the tape in dot format and you
        can render it using Graphviz dot.

        Args:
            filename (str): File to save the visualisation. Default None.
        """
        G = self.create_graph()
        from networkx.drawing.nx_agraph import write_dot
        write_dot(G, filename)

    def visualise_pdf(self, filename):
        """Create a PDF visualisation of the tape.

        This depends on the Python package networkx and the external Graphviz
        package. The latter can be installed using e.g.::

            sudo apt install graphviz

        on Ubuntu or::

            brew install graphviz

        on Mac.

        Args:
            filename (str): File to save the visualisation. Must end in .pdf.
        """
        if not filename.endswith(".pdf"):
            raise ValueError("Filename for PDF output must end in .pdf")
        from networkx.drawing.nx_agraph import to_agraph
        A = to_agraph(self.create_graph())
        A.draw(filename, prog="dot")

    @property
    def progress_bar(self):
        """Specify a progress bar class to print during tape evaluation.

        Setting this attribute to a subclass of :class:`progress.bar.Bar` will
        cause every evaluation of a reduced functional, adjoint, TLM or Hessian
        to print a progress bar.

        For example, the following code::

            from progress.bar import FillingSquaresBar
            tape = get_working_tape()
            tape.progress_bar = FillingSquaresBar

        will cause tape evaluations to print progress bars similar to the
        following::

            Evaluating functional ▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣ 100%
            Evaluating adjoint ▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣▣ 100%

        For information on available progress bar styles and their
        configuration, see the `progress package documentation
        <https://pypi.org/project/progress/>`_.
        """
        return self._bar

    @progress_bar.setter
    def progress_bar(self, bar):
        self._bar = bar


class _NullProgressBar:
    """A placeholder class with the same interface as a progress bar."""

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        pass

    def iter(self, iterator):
        return iterator


class TapeTimeStepper:
    """Iterator wrapper which advances the timestep after each iteration."""
    def __init__(self, tape, iterable):
        self.tape = tape
        self.iterator = tape.progress_bar("Taping forward").iter(iterable)
        self._first = True

    def __iter__(self):
        return self

    def __next__(self):
        step = next(self.iterator)
        if self._first:
            self._first = False
        else:
            self.tape.end_timestep()
        return step


class TimeStep(list):
    """A list of blocks in a single time step, plus associated metadata."""
    def __init__(self, blocks=()):
        super().__init__(blocks)
        # The set of block variables which are needed to restart from the start
        # of this timestep.
        self.checkpointable_state = OrderedSet()
        self.adjoint_dependencies = OrderedSet()
        # A dictionary mapping the block variables in the checkpointable state
        # to their checkpoint values.
        self._checkpoint = {}
        # Flag indicating whether the adjoint dependencies have been revised
        # by removing outputs not marked in the path and adding checkpointable
        # states that are marked in the path.
        self._revised_adj_deps = False

    def copy(self, blocks=None):
        out = TimeStep(blocks or self)
        out.checkpointable_state = self.checkpointable_state
        return out

    def checkpoint(self, checkpointable_state, adj_dependencies, global_deps):
        """Store a copy of the checkpoints in the checkpointable state.

        Args:
            checkpointable_state (bool): If True, store the checkpointable state
            required to restart from the start of a timestep.
            adj_dependencies (bool): If True, store the adjoint dependencies required
            to compute the adjoint of a timestep.
            global_deps (set): This set stores the common dependencies for all timesteps.
            For additional information, please refer to the :class:`CheckpointManager`
            documentation.
        """
        with stop_annotating():
            if checkpointable_state:
                for var in self.checkpointable_state:
                    if var in global_deps:
                        # Creating a new checkpoint object is not necessary here
                        # because the global dependencies do not change.
                        self._checkpoint[var] = var._checkpoint
                    else:
                        self._checkpoint[var] = var.saved_output._ad_create_checkpoint()

            if adj_dependencies:
                if self._revised_adj_deps:
                    for var in self.adjoint_dependencies:
                        self._checkpoint[var] = var.saved_output._ad_create_checkpoint()
                else:
                    # The adjoint dependencies have not been revised yet. At this stage,
                    # the block nodes are not marked in the path because the control variable(s)
                    # are not yet determined.
                    for var in self.adjoint_dependencies.union(self.checkpointable_state):
                        self._checkpoint[var] = var.saved_output._ad_create_checkpoint()

    def restore_from_checkpoint(self, from_storage):
        """Restore the block var checkpoints from the timestep checkpoint."""
        from .overloaded_type import OverloadedType
        for var, checkpoint in self._checkpoint.items():
            if (
                from_storage == StorageType.DISK
                and isinstance(checkpoint, OverloadedType)
            ):
                # checkpoint._ad_restore_checkpoint should be able to restore
                # from disk.
                var.checkpoint = checkpoint._ad_restore_at_checkpoint(checkpoint)
            else:
                var.checkpoint = checkpoint

    def delete_checkpoint(self):
        """Delete the stored checkpoint references."""
        self._checkpoint = {}


class TimeStepSequence(list):
    """A list of Blocks separated into timesteps to facilitate checkpointing.

    This behaves like a list of blocks. To access a list of the timesteps, use
    the :attr:`steps` property.

    Args:
        blocks (list[Block] or TimeStepSequence, optional): If provided, `blocks` can be a list of :class:`Block`
        or another TimeStepSequence to copy from.
        steps (list[TimeStep], optional): If provided, `steps` should be a list of :class:`TimeStep` to copy from.

    Attributes:
        steps (list[TimeStep]): A list of timesteps.
    """

    def __init__(self, blocks=None, steps: Optional[Iterable[Iterable[TimeStep]]] = None):
        # Keep both per-timestep and unified block lists.
        if steps and blocks:
            raise ValueError("set blocks or steps but not both.")
        elif isinstance(blocks, TimeStepSequence):
            self._steps = [step.copy() for step in blocks._steps]
        elif blocks:
            self._steps = [TimeStep(blocks)]
        else:
            self._steps = list(step.copy() for step in steps) if steps else []
        super().__init__(chain.from_iterable(self._steps))

    @property
    def steps(self):
        return self._steps

    def append(self, other):
        """Add a new block to the sequence and to the current TimeStep."""
        if not self.steps:
            self.append_step()
        self._steps[-1].append(other)
        super().append(other)

    def append_step(self, step=None):
        """Add a new TimeStep."""
        self._steps.append(step or TimeStep())

    def __setitem__(self, key, value):
        raise ValueError(
            "Unable to set arbitrary blocks. Try appending instead."
        )

    def __delitem__(self, key, value):
        raise ValueError(
            "Unable to delete blocks from sequence."
        )


class TapePackageData(ABC):
    """Abstract base class for additional data that packages store on the tape.

    If a package that uses Pyadjoint needs to store additional tape state, such
    as the location of checkpoint files, it should store an instance of a
    subclass of `TapePackageData` in the :attr:`_package_data` dictionary of
    the tape. E.g::

        get_working_tape()._package_data["firedrake"] = checkpoint_data
    """

    @abstractmethod
    def clear(self):
        """Delete the current state of the object.

        This is called when the tape is cleared."""
        pass

    @abstractmethod
    def reset(self):
        """Reset the current state before a new forward evaluation."""
        pass

    @abstractmethod
    def checkpoint(self):
        """Record the information required to return to the current state."""
        pass

    @abstractmethod
    def restore_from_checkpoint(self, state):
        """Restore state from a previously stored checkpoint."""
        pass

    @abstractmethod
    def copy(self):
        """Produce a new copy of state to be passed to a copy of the tape."""
        pass

    @abstractmethod
    def continue_checkpointing(self):
        """Continue the checkpointing process on disk.
        """
        pass

    @abstractmethod
    def pause_checkpointing(self):
        """Pause the checkpointing process on disk.
        """
        pass



import logging

from .enlisting import Enlist
from .tape import stop_annotating


def taylor_test(J, m, h, dJdm=None, Hm=0):
    """Run a taylor test on the functional J around point m in direction h.

    Given a functional J, a point in control space m, and a direction in
    control space h, the function computes the taylor remainders and
    returns the convergence rate.

    Args:
        J (reduced_functional.ReducedFunctional): The functional to evaluate the taylor remainders of.
            Must be an instance of :class:`ReducedFunctional`, or something with a similar
            interface.
        m (overloaded_type.OverloadedType): The expansion points in control space. Must be of same type as the
            control.
        h (overloaded_type.OverloadedType): The direction of perturbations. Must be of same type as
            the control.

    Returns:
        float: The smallest computed convergence rate of the tested perturbations.

    """
    with stop_annotating():
        hs = Enlist(h)
        ms = Enlist(m)

        if len(hs) != len(ms):
            raise ValueError(
                "%d perturbations are given but only %d expansion points are provided" % (len(hs), len(ms)))

        Jm = J(m)
        if dJdm is None:
            ds = Enlist(J.derivative())
            if len(ds) != len(ms):
                raise ValueError(
                    "The derivative of J depends on %d variables but only %d expansion points are given" % (
                        len(ds), len(ms)))
            dJdm = sum(hi._ad_dot(di) for hi, di in zip(hs, ds))

        def perturbe(eps):
            ret = [mi._ad_add(hi._ad_mul(eps)) for mi, hi in zip(ms, hs)]
            return ms.delist(ret)

        print("Running Taylor test")
        residuals = []
        epsilons = [0.01 / 2 ** i for i in range(4)]
        for eps in epsilons:
            Jp = J(perturbe(eps))
            res = abs(Jp - Jm - eps * dJdm - 0.5 * eps ** 2 * Hm)
            residuals.append(res)

        if min(residuals) < 1E-15:
            logging.warning("The taylor remainder is close to machine precision.")
        print("Computed residuals: {}".format(residuals))
    return min(convergence_rates(residuals, epsilons))


def convergence_rates(E_values, eps_values, show=True):
    from numpy import log
    r = []
    for i in range(1, len(eps_values)):
        r.append(log(E_values[i] / E_values[i - 1])
                 / log(eps_values[i] / eps_values[i - 1]))
    if show:
        print("Computed convergence rates: {}".format(r))
    return r


def taylor_to_dict(J, m, h):
    """Run a 0th, 1st and second order taylor test on the functional J
      around point m in direction h.

    Given a functional J, a point in control space m, and a direction in
    control space h, the function computes the taylor remainders and
    returns the convergence rate in a dictionary.

    Args:
        J (reduced_functional.ReducedFunctional): The functional to evaluate the taylor remainders of.
            Must be an instance of :class:`ReducedFunctional`, or something with a similar
            interface.
        m (overloaded_type.OverloadedType): The expansion points in control space. Must be of same type as the
            control.
        h (overloaded_type.OverloadedType): The direction of perturbations. Must be of same type as
            the control.

    Returns:
        dict: The perturbation sizes, residuals and rates of the tests.

            eps (list): List of all perturbation sizes used, eps[i]*h.
            R0 (dict): Results from 0th order taylor test (finite difference).
                Residual (list): The computed residuals.
                Rate (list): The computed convergence rates based on eps and residuals. Expected to be 1.0.
            R1 (dict): Results from the 1st order taylor test.
                Residual (list): The computed residuals.
                Rate (list): The computed convergence rates based on eps and residuals. Expected to be 2.0.
            R2 (dict): Results from the 2nd order taylor test.
                Residual (list): The computed residuals.
                Rate (list): The computed convergence rates based on eps and residuals. Expected to be 3.0.

    """
    with stop_annotating():
        hs = Enlist(h)
        ms = Enlist(m)

        if len(hs) != len(ms):
            raise ValueError("{0:d} perturbations are given but only {1:d} expansion points are provided"
                             .format(len(hs), len(ms)))

        Jm = J(m)
        print("Computing derivative")
        ds = Enlist(J.derivative())
        if len(ds) != len(ms):
            raise ValueError("The derivative of J depends on {0:d} variables"
                             .format(len(ds))
                             + "but only {0:d} expansion points are given".
                             format(len(ms)))
        dJdm = sum(hi._ad_dot(di) for hi, di in zip(hs, ds))

        print("Computing Hessian")

        Hm = Enlist(J.hessian(hs))
        Hmh = sum(hi._ad_dot(hmi) for hi, hmi in zip(hs, Hm))

        def perturbe(eps):
            ret = [mi._ad_add(hi._ad_mul(eps)) for mi, hi in zip(ms, hs)]
            return ms.delist(ret)

        print("Running Taylor test")
        error_dict = {"eps": None, "R0": {"Residual": [], "Rate": None},
                      "R1": {"Residual": [], "Rate": None},
                      "R2": {"Residual": [], "Rate": None}}

        epsilons = [0.01 / 2**i for i in range(4)]
        error_dict["eps"] = epsilons
        for eps in epsilons:
            Jp = J(perturbe(eps))
            error_dict["R0"]["Residual"].append(abs(Jp - Jm))
            error_dict["R1"]["Residual"].append(abs(Jp - Jm - eps * dJdm))
            error_dict["R2"]["Residual"].append(abs(Jp - Jm - eps * dJdm - 0.5 * eps**2 * Hmh))

        for key in error_dict.keys():
            if key != "eps":
                error_dict[key]["Rate"] = convergence_rates(error_dict[key]
                                                            ["Residual"][:],
                                                            error_dict["eps"],
                                                            show=False)
    J(m)
    return error_dict



import numpy as np

from ..reduced_functional import ReducedFunctional
from ..reduced_functional_numpy import ReducedFunctionalNumPy, gather
from ..tape import no_annotations


class SciPyConvergenceError(Exception):
    """Raised if a SciPy minimization fails.
    """


def serialise_bounds(rf_np, bounds):
    """ Converts bounds to an array of (min, max) tuples and serialises it in a parallel environment. """

    if len(np.array(bounds).shape) == 1:
        bounds = np.array([[b] for b in bounds])

    if len(bounds) != 2:
        raise ValueError(
            "The 'bounds' parameter must be of the form [lower_bound, upper_bound] for one parameter"
            "or [ [lower_bound1, lower_bound2, ...], [upper_bound1, upper_bound2, ...] ] for multiple parameters.")

    bounds_arr = [[], []]
    for i in range(2):
        for j in range(len(bounds[i])):
            bound = bounds[i][j]
            if bound is None or isinstance(bound, (int, float)):
                bound_len = len(rf_np.get_global(rf_np.controls[j]))
                const_bound = np.array([bound] * bound_len)

                bounds_arr[i] += const_bound.tolist()
            else:
                bounds_arr[i] += rf_np.obj_to_array(bound).tolist()

    # Transpose and return the array to get the form
    # [ [lower_bound1, upper_bound1], [lower_bound2, upper_bound2], ... ]
    return np.array(bounds_arr).T


def minimize_scipy_generic(rf_np, method, bounds=None, **kwargs):
    """Interface to the generic minimize method in scipy

    """
    try:
        from scipy.optimize import minimize as scipy_minimize
    except ImportError:
        print("**************** Deprecated warning *****************")
        print("You have an unusable installation of scipy. This version is not supported by dolfin-adjoint.")

        try:
            import scipy
            print("Version: %s\tFile: %s" % (scipy.__version__, scipy.__file__))
        except ImportError:
            pass

        raise

    m = [p.tape_value() for p in rf_np.controls]
    m_global = rf_np.obj_to_array(m)
    J = rf_np.__call__
    dJ = lambda m: rf_np.derivative()
    H = lambda x, p: rf_np.hessian(p)

    if "options" not in kwargs:
        kwargs["options"] = {}
        # TODO: What to do here?
        """
    if rank(rf_np.rf.mpi_comm()) != 0:
        # Shut up all processors except the first one.
        kwargs["options"]["disp"] = False
        """
    else:
        # Print out progress information by default
        if "disp" not in kwargs["options"]:
            kwargs["options"]["disp"] = True

    # Make the default SLSLQP options more verbose
    if method == "SLSQP" and "iprint" not in kwargs["options"]:
        kwargs["options"]["iprint"] = 2

    # For gradient-based methods add the derivative function to the argument list
    if method not in ["COBYLA", "Nelder-Mead", "Anneal", "Powell"]:
        kwargs["jac"] = dJ

    # For Hessian-based methods add the Hessian action function to the argument list
    if method in ["Newton-CG"]:
        kwargs["hessp"] = H

    if "constraints" in kwargs:
        from .constraints import canonicalise, InequalityConstraint, EqualityConstraint
        constraints = canonicalise(kwargs["constraints"])
        scipy_c = []
        for c in constraints:
            if isinstance(c, InequalityConstraint):
                typestr = "ineq"
            elif isinstance(c, EqualityConstraint):
                typestr = "eq"
            else:
                raise Exception("Unknown constraint class")

            def jac(x):
                out = c.jacobian(x)
                return [gather(y) for y in out]

            scipy_c.append(dict(type=typestr, fun=c.function, jac=jac))
        kwargs["constraints"] = scipy_c

    if method == "basinhopping":
        try:
            from scipy.optimize import basinhopping
        except ImportError:
            print("**************** Outdated scipy version warning *****************")
            print("The basin hopping optimisation algorithm requires scipy >= 0.12.")
            raise ImportError

        del kwargs["options"]
        del kwargs["jac"]
        kwargs["minimizer_kwargs"]["jac"] = dJ

        if "bounds" in kwargs["minimizer_kwargs"]:
            kwargs["minimizer_kwargs"]["bounds"] = \
                serialise_bounds(rf_np, kwargs["minimizer_kwargs"]["bounds"])

        res = basinhopping(J, m_global, **kwargs)

    elif bounds is not None:
        bounds = serialise_bounds(rf_np, bounds)
        res = scipy_minimize(J, m_global, method=method, bounds=bounds, **kwargs)
    else:
        res = scipy_minimize(J, m_global, method=method, **kwargs)

    if not res.success:
        raise SciPyConvergenceError(f"SciPy minimization failed because: {res.message}")

    m = rf_np.set_controls(np.array(res["x"]))
    return m


def minimize_custom(rf_np, bounds=None, **kwargs):
    """ Interface to the user-provided minimisation method """

    try:
        algo = kwargs["algorithm"]
        del kwargs["algorithm"]
    except KeyError:
        raise KeyError(
            'When using a "Custom" optimisation method, you must pass the optimisation function as '
            'the "algorithm" parameter. Make sure that this function accepts the same arguments as '
            'scipy.optimize.minimize.')

    m = [p.tape_value() for p in rf_np.controls]
    m_global = rf_np.obj_to_array(m)
    J = rf_np.__call__

    dJ = lambda m: rf_np.derivative(m)
    H = rf_np.hessian

    if bounds is not None:
        bounds = serialise_bounds(rf_np, bounds)

    res = algo(J, m_global, dJ, H, bounds, **kwargs)

    try:
        m = rf_np.set_controls(np.array(res))
    except Exception as e:
        # TODO: Is this possible?
        raise e(
            "Failed to update the optimised control values. "
            "Are you sure your custom optimisation algorithm returns an array containing the optimised values?")
    return m


optimization_algorithms_dict = {'L-BFGS-B': ('The L-BFGS-B implementation in scipy.', minimize_scipy_generic),
                                'SLSQP': ('The SLSQP implementation in scipy.', minimize_scipy_generic),
                                'TNC': (
                                'The truncated Newton algorithm implemented in scipy.', minimize_scipy_generic),
                                'CG': ('The nonlinear conjugate gradient algorithm implemented in scipy.',
                                       minimize_scipy_generic),
                                'BFGS': ('The BFGS implementation in scipy.', minimize_scipy_generic),
                                'Nelder-Mead': ('Gradient-free Simplex algorithm.', minimize_scipy_generic),
                                'Powell': ('Gradient-free Powells method', minimize_scipy_generic),
                                'Newton-CG': ('Newton-CG method', minimize_scipy_generic),
                                'Anneal': ('Gradient-free simulated annealing', minimize_scipy_generic),
                                'basinhopping': ('Global basin hopping method', minimize_scipy_generic),
                                'COBYLA': ('Gradient-free constrained optimization by linear approxition method',
                                           minimize_scipy_generic),
                                'Custom': ('User-provided optimization algorithm', minimize_custom)
                                }


def print_optimization_methods():
    """ Prints the available optimization methods """

    print('Available optimization methods:')
    for function_name, (description, func) in optimization_algorithms_dict.items():
        print(function_name, ': ', description)


@no_annotations
def minimize(rf, method='L-BFGS-B', scale=1.0, **kwargs):
    """Solves the minimisation problem with PDE constraint:

           min_m func(u, m)
             s.t.
           e(u, m) = 0
           lb <= m <= ub
           g(m) <= u

        where m is the control variable, u is the solution of the PDE system e(u, m) = 0, func is the functional
        of interest and lb, ub and g(m) constraints the control variables.
        The optimization problem is solved using a gradient based optimization algorithm and the functional gradients
        are computed by solving the associated adjoint system.

        The function arguments are as follows:
        * 'rf' must be a ReducedFunctional object.
        * 'method' specifies the optimization method to be used to solve the problem. The available methods can be
            listed with the print_optimization_methods function.
        * 'scale' is a factor to scale to problem (default: 1.0).
        * 'bounds' is an optional keyword parameter to support control constraints: bounds = (lb, ub).
            lb and ub must be of the same type than the parameters m.

        Additional arguments specific for the optimization algorithms can be added to the minimize functions
        (e.g. iprint = 2). These arguments will be passed to the underlying optimization algorithm.
        For detailed information about which arguments are supported for each optimization algorithm,
        please refer to the documentaton of the optimization algorithm.

    """
    rf.scale = scale
    if isinstance(rf, ReducedFunctionalNumPy):
        rf_np = rf
    elif isinstance(rf, ReducedFunctional):
        rf_np = ReducedFunctionalNumPy(rf)
    else:
        # Assume the user knows what he is doing - he might for example written
        # his own reduced functional class.
        rf_np = rf

    try:
        algorithm = optimization_algorithms_dict[method][1]
    except KeyError:
        raise KeyError(
            'Unknown optimization method '
            + method
            + '. Use print_optimization_methods() to get a list of the available methods.')

    if algorithm == minimize_scipy_generic:
        # For scipy's generic inteface we need to pass the optimisation method as a parameter.
        kwargs["method"] = method

    opt = algorithm(rf_np, **kwargs)

    if len(opt) == 1:
        return opt[0]
    else:
        return opt


def maximize(rf, method='L-BFGS-B', scale=1.0, **kwargs):
    """ Solves the maximisation problem with PDE constraint:

           max_m func(u, m)
             s.t.
           e(u, m) = 0
           lb <= m <= ub
           g(m) <= u

        where m is the control variable, u is the solution of the PDE system e(u, m) = 0,
        func is the functional of interest and lb, ub and g(m) constraints the control variables.
        The optimization problem is solved using a gradient based optimization algorithm
        and the functional gradients are computed by solving the associated adjoint system.

        The function arguments are as follows:
        * 'rf' must be a ReducedFunctional object.
        * 'method' specifies the optimization method to be used to solve the problem.
            The available methods can be listed with the print_optimization_methods function.
        * 'scale' is a factor to scale to problem (default: 1.0).
        * 'bounds' is an optional keyword parameter to support control constraints: bounds = (lb, ub).
            lb and ub must be of the same type than the parameters m.

        Additional arguments specific for the optimization methods can be added to the minimize functions
        (e.g. iprint = 2). These arguments will be passed to the underlying optimization method.
        For detailed information about which arguments are supported for each optimization method,
        please refer to the documentaton of the optimization algorithm.
        """
    return minimize(rf, method, scale=-scale, **kwargs)


minimise = minimize
maximise = maximize



from . import optimization_problem


class OptimizationSolver(object):
    """An abstract base class that represents an optimization solver."""
    def __init__(self, problem, parameters=None):
        self.__check_arguments(problem, parameters)

        #: problem: an OptimizationProblem instance.
        self.problem = problem

        #: parameters: a dictionary of parameters.
        self.parameters = parameters

    def __check_arguments(self, problem, parameters):
        if not isinstance(problem, optimization_problem.OptimizationProblem):
            raise TypeError("problem should be an OptimizationProblem.")

    def solve(self):
        raise NotImplementedError("This class is abstract.")



from functools import cached_property, lru_cache, wraps
from itertools import count
import numbers
import operator
from .block import Block
from .overloaded_type import OverloadedType, register_overloaded_type
from .tape import get_working_tape, annotate_tape
import numpy as np
import sympy as sp

__all__ = ["AdjFloat"]

_op_fns = {}


def register_function(np_operator):
    def register(cls):
        _op_fns[cls.__name__] = np_operator
        return cls
    return register


@register_function(np.power)
class _pyadjoint_power(sp.Function):
    def fdiff(self, argindex=1):
        if argindex == 1:
            return sp.Piecewise(
                # Let SymPy decide how to handle indeterminate form
                ((self.args[0] ** self.args[1]).diff(self.args[0]), sp.And(self.args[0] == 0, self.args[1] == 0)),
                # Otherwise simplify
                (sp.S.Zero, self.args[1] == 0),
                (_pyadjoint_power(self.args[0], self.args[1] - 1) * self.args[1], True))
        elif argindex == 2:
            return (self.args[0] ** self.args[1]).diff(self.args[1])


@register_function(np.hypot)
class _pyadjoint_hypot(sp.Function):
    def fdiff(self, argindex=1):
        if argindex == 1:
            return self.args[0] / _pyadjoint_hypot(self.args[0],
                                                   self.args[1])
        elif argindex == 2:
            return self.args[1] / _pyadjoint_hypot(self.args[0],
                                                   self.args[1])


@register_function(np.expm1)
class _pyadjoint_expm1(sp.Function):
    def fdiff(self, argindex=1):
        if argindex == 1:
            return sp.exp(self.args[0])


@register_function(np.log1p)
class _pyadjoint_log1p(sp.Function):
    def fdiff(self, argindex=1):
        if argindex == 1:
            return sp.Integer(1) / (sp.Integer(1) + self.args[0])


@lru_cache(maxsize=256)
def codegen(expr, symbols, diff=()):
    for idx in diff:
        expr = expr.diff(symbols[idx])
    return sp.lambdify(symbols, expr, modules=["numpy", _op_fns])


class Operator:
    _symbol_count = count()

    def __init__(self, sp_operator, nargs):
        self._sp_operator = sp_operator
        self._nargs = nargs

    @property
    def sp_operator(self):
        return self._sp_operator

    @property
    def nargs(self):
        return self._nargs

    @cached_property
    def symbols(self):
        return tuple(sp.Symbol(f"_pyadjoint_symbol_{next(self._symbol_count)}", real=True) for _ in range(self.nargs))

    @cached_property
    def expr(self):
        return self.sp_operator(*self.symbols)

    def codegen(self, diff=()):
        return codegen(self.expr, self.symbols, diff=diff)


class AdjFloatExprBlock(Block):
    def __init__(self, operator, *args, np_operator=None):
        super().__init__()
        self._operator = operator
        self._np_operator = np_operator
        for arg in args:
            self.add_dependency(arg)

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_input, = adj_inputs
        return self._operator.codegen(diff=(idx,))(*inputs) * adj_input

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        if idx != 0:
            raise ValueError("Unexpected idx")
        val = 0.0
        for idx1 in range(self._operator.nargs):
            if tlm_inputs[idx1] is not None:
                val += self._operator.codegen(diff=(idx1,))(*inputs) * tlm_inputs[idx1]
        return val

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        hessian_input, = hessian_inputs
        adj_input, = adj_inputs
        val = self._operator.codegen(diff=(idx,))(*inputs) * hessian_input
        for idx1, dep in relevant_dependencies:
            tlm_input = dep.tlm_value
            if tlm_input is not None:
                val += self._operator.codegen(diff=(idx, idx1))(*inputs) * adj_input * tlm_input
        return val

    def recompute_component(self, inputs, block_variable, idx, prepared):
        if idx != 0:
            raise ValueError("Unexpected idx")
        if self._np_operator is None:
            return self._operator.codegen(diff=())(*inputs)
        else:
            return self._np_operator(*inputs)


def annotate_operator(operator, np_operator=None):
    def wrapper(fn):
        @wraps(fn)
        def annotated_operator(*args):
            output = fn(*args)
            if not isinstance(output, numbers.Complex):
                # Not annotated
                return output
            output = AdjFloat(output)  # Error here if not real

            if annotate_tape():
                args = list(args)
                for i, arg in enumerate(args):
                    if isinstance(arg, OverloadedType):
                        pass
                    elif isinstance(arg, numbers.Complex):
                        args[i] = AdjFloat(arg)  # Error here if not real
                    else:
                        # Not annotated
                        return output

                block = AdjFloatExprBlock(
                    operator, *args,
                    np_operator=fn if np_operator is None else np_operator)
                tape = get_working_tape()
                tape.add_block(block)
                block.add_output(output.block_variable)
            return output
        return annotated_operator
    return wrapper


def roperator(operator):
    def roperator(a, b):
        return operator(b, a)
    return roperator


_ops = {}


def register_operator(np_operator, sp_operator, nargs):
    @annotate_operator(Operator(sp_operator, nargs))
    def wrapped_operator(*args):
        if len(args) != nargs:
            return NotImplemented
        return np_operator(*(float(arg) if isinstance(arg, AdjFloat) else arg for arg in args))
    _ops[np_operator] = wrapped_operator
    return _ops[np_operator]


@register_overloaded_type
class AdjFloat(OverloadedType, float):
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc not in _ops:
            # Not annotated
            return getattr(ufunc, method)(
                *(float(arg) if isinstance(arg, AdjFloat) else arg for arg in inputs), **kwargs)
        if method != "__call__":
            return NotImplemented
        if len(kwargs) > 0:
            return NotImplemented
        return _ops[ufunc](*inputs)

    @annotate_operator(Operator(lambda x: sp.Piecewise((x, x >= 0), (-x, True)), 1), operator.abs)
    def __abs__(self):
        return super().__abs__()

    @annotate_operator(Operator(operator.pos, 1), operator.pos)
    def __pos__(self):
        return super().__pos__()

    @annotate_operator(Operator(operator.neg, 1), operator.neg)
    def __neg__(self):
        return super().__neg__()

    @annotate_operator(Operator(operator.mul, 2), operator.mul)
    def __mul__(self, other):
        return super().__mul__(other)

    @annotate_operator(Operator(roperator(operator.mul), 2), roperator(operator.mul))
    def __rmul__(self, other):
        return super().__rmul__(other)

    @annotate_operator(Operator(operator.truediv, 2), operator.truediv)
    def __truediv__(self, other):
        return super().__truediv__(other)

    @annotate_operator(Operator(roperator(operator.truediv), 2), roperator(operator.truediv))
    def __rtruediv__(self, other):
        return super().__rtruediv__(other)

    @annotate_operator(Operator(operator.add, 2), operator.add)
    def __add__(self, other):
        return super().__add__(other)

    @annotate_operator(Operator(roperator(operator.add), 2), roperator(operator.add))
    def __radd__(self, other):
        return super().__radd__(other)

    @annotate_operator(Operator(operator.sub, 2), operator.sub)
    def __sub__(self, other):
        return super().__sub__(other)

    @annotate_operator(Operator(roperator(operator.sub), 2), roperator(operator.sub))
    def __rsub__(self, other):
        return super().__rsub__(other)

    @annotate_operator(Operator(_pyadjoint_power, 2), operator.pow)
    def __pow__(self, other):
        return super().__pow__(other)

    @annotate_operator(Operator(roperator(_pyadjoint_power), 2), roperator(operator.pow))
    def __rpow__(self, other):
        return super().__rpow__(other)

    absolute = register_operator(np.absolute, lambda x: sp.Piecewise((x, x >= 0), (-x, True)), 1)
    positive = register_operator(np.positive, operator.pos, 1)
    negative = register_operator(np.negative, operator.neg, 1)
    multiply = register_operator(np.multiply, operator.mul, 2)
    divide = register_operator(np.divide, operator.truediv, 2)
    add = register_operator(np.add, operator.add, 2)
    subtract = register_operator(np.subtract, operator.sub, 2)
    power = register_operator(np.power, _pyadjoint_power, 2)
    minimum = register_operator(
        np.minimum,
        lambda self, other: sp.Piecewise((self, self <= other),
                                         (other, True)),
        2)
    maximum = register_operator(
        np.maximum,
        lambda self, other: sp.Piecewise((self, self >= other),
                                         (other, True)),
        2)

    sin = register_operator(np.sin, sp.sin, 1)
    cos = register_operator(np.cos, sp.cos, 1)
    tan = register_operator(np.tan, sp.tan, 1)
    arcsin = register_operator(np.arcsin, sp.asin, 1)
    arccos = register_operator(np.arccos, sp.acos, 1)
    arctan = register_operator(np.arctan, sp.atan, 1)
    arctan2 = register_operator(np.arctan2, sp.atan2, 2)
    hypot = register_operator(np.hypot, _pyadjoint_hypot, 2)
    sinh = register_operator(np.sinh, sp.sinh, 1)
    cosh = register_operator(np.cosh, sp.cosh, 1)
    tanh = register_operator(np.tanh, sp.tanh, 1)
    arcsinh = register_operator(np.arcsinh, sp.asinh, 1)
    arccosh = register_operator(np.arccosh, sp.acosh, 1)
    arctanh = register_operator(np.arctanh, sp.atanh, 1)
    exp = register_operator(np.exp, sp.exp, 1)
    exp2 = register_operator(np.exp2, lambda x: 2 ** x, 1)
    expm1 = register_operator(np.expm1, _pyadjoint_expm1, 1)
    log = register_operator(np.log, sp.log, 1)
    log2 = register_operator(np.log2, lambda x: sp.log(x, 2), 1)
    log10 = register_operator(np.log10, lambda x: sp.log(x, 10), 1)
    log1p = register_operator(np.log1p, _pyadjoint_log1p, 1)
    sqrt = register_operator(np.sqrt, sp.sqrt, 1)
    square = register_operator(np.square, lambda x: x ** 2, 1)
    cbrt = register_operator(np.cbrt, lambda x: x ** sp.Rational(1, 3), 1)
    reciprocal = register_operator(np.reciprocal, lambda x: sp.Integer(1) / x, 1)

    def _ad_init_zero(self, dual=False):
        return type(self)(0.)

    def _ad_convert_riesz(self, value, riesz_map=None):
        if riesz_map is not None:
            raise ValueError(f"Unexpected Riesz map: {riesz_map}")
        return type(self)(value)

    def _ad_create_checkpoint(self):
        # Floats are immutable.
        return self

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def _ad_mul(self, other):
        return super().__mul__(other)

    def _ad_add(self, other):
        return super().__add__(other)

    def _ad_dot(self, other):
        return super().__mul__(other)

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        dst = type(dst)(src[offset])
        offset += 1
        return dst, offset

    @staticmethod
    def _ad_to_list(value):
        return [value]

    def _ad_copy(self):
        return self

    @property
    def _ad_str(self):
        """Return the string of the taped value of this variable."""
        return str(self.block_variable.saved_output)

    def _ad_to_petsc(self, vec=None):
        raise NotImplementedError("_ad_to_petsc not implemented for AdjFloat.")

    def _ad_from_petsc(self, vec):
        raise NotImplementedError("_ad_from_petsc not implemented for AdjFloat.")


# Backwards compatibility
exp = np.exp
log = np.log
min = np.minimum
max = np.maximum



from __future__ import print_function
from .reduced_functional import AbstractReducedFunctional, ReducedFunctional
from .tape import no_annotations, get_working_tape
from .enlisting import Enlist
from .control import Control
from .adjfloat import AdjFloat

import numpy


class ReducedFunctionalNumPy(AbstractReducedFunctional):
    """This class implements the reduced functional for given functional and
    controls based on numpy data structures.

    This "NumPy version" of the pyadjoint.ReducedFunctional is created from
    an existing ReducedFunctional object:
    rf_np = ReducedFunctionalNumPy(rf = rf)
    """

    def __init__(self, functional, controls=None, tape=None):
        if isinstance(functional, AdjFloat):
            functional = ReducedFunctional(functional=functional,
                                           controls=controls,
                                           tape=tape)
        self.rf = functional

    @property
    def controls(self) -> list[Control]:
        return self.rf._controls

    def __getattr__(self, item):
        return getattr(self.rf, item)

    def __call__(self, m_array):
        """An implementation of the reduced functional evaluation
            that accepts the control values as an array of scalars

        """
        m_copies = [control.copy_data() for control in self.controls]
        return self.rf.__call__(self.set_local(m_copies, m_array))

    def set_local(self, m, m_array):
        offset = 0
        for i, control in enumerate(self.controls):
            m[i], offset = control.assign_numpy(m[i], m_array, offset)

        return m

    def get_global(self, m):
        m_global = []
        for i, v in enumerate(Enlist(m)):
            if isinstance(v, Control):
                # TODO: Consider if you want this design.
                m_global += v.fetch_numpy(v.control)
            elif hasattr(v, "_ad_to_list"):
                m_global += v._ad_to_list(v)
            else:
                m_global += self.controls[i].control._ad_to_list(v)
        return numpy.array(m_global, dtype="d")

    @no_annotations
    def derivative(self, adj_input=1.0, apply_riesz=True):

        if not apply_riesz:
            raise ValueError(
                "ReducedFunctionalNumpy only returns primal gradients."
            )

        dJdm = self.rf.derivative(adj_input=adj_input,
                                  apply_riesz=apply_riesz)
        dJdm = Enlist(dJdm)

        m_global = []
        for i, control in enumerate(self.controls):
            # This is a little ugly, but we need to go through the control to get to the OverloadedType.
            # There is no guarantee that dJdm[i] is an OverloadedType and not a backend type.
            m_global += control.fetch_numpy(dJdm[i])

        return numpy.array(m_global, dtype="d")

    @no_annotations
    def hessian(self, m_dot_array, apply_riesz=True):

        if not apply_riesz:
            raise ValueError(
                "ReducedFunctionalNumpy only returns primal gradients."
            )

        # Calling derivative is needed, see i.e. examples/stokes-shape-opt
        self.derivative()
        m_copies = [control.copy_data() for control in self.controls]
        Hm = self.rf.hessian(self.set_local(m_copies, m_dot_array),
                             apply_riesz=apply_riesz)
        Hm = Enlist(Hm)

        m_global = []
        for i, control in enumerate(self.controls):
            # This is a little ugly, but we need to go through the control to get to the OverloadedType.
            # There is no guarantee that dJdm[i] is an OverloadedType and not a backend type.
            m_global += control.fetch_numpy(Hm[i])

        tape = get_working_tape()
        tape.reset_variables()

        return numpy.array(m_global, dtype="d")

    @no_annotations
    def tlm(self, m_dot_array):
        m_copies = [control.copy_data() for control in self.controls]
        tm = self.rf.tlm(self.set_local(m_copies, m_dot_array))
        tm_global = self.functional._ad_to_list(tm)

        tape = get_working_tape()
        tape.reset_variables()

        return numpy.array(tm_global, dtype="d")

    def obj_to_array(self, obj):
        return self.get_global(obj)

    def get_controls(self):
        m = [p.tape_value() for p in self.controls]
        return self.obj_to_array(m)

    def set_controls(self, array):
        m = [p.tape_value() for p in self.controls]
        m = self.set_local(m, array)
        for control, m_i in zip(self.controls, m):
            control.update(m_i)
        return m


def set_local(coeffs, m_array):
    offset = 0
    for m in Enlist(coeffs):
        _, offset = m._ad_assign_numpy(m, m_array, offset)

    return coeffs


def gather(m):
    if isinstance(m, list):
        return list(map(gather, m))
    elif hasattr(m, "_ad_to_list"):
        return m._ad_to_list(m)
    else:
        return m  # Assume it is gathered already



# flake8: noqa
from .optimization.moola_problem import MoolaOptimizationProblem
from .optimization.constraints import InequalityConstraint, EqualityConstraint
from .optimization.tao_solver import TAOSolver
from .optimization.rol_solver import ROLSolver
from .optimization.ipopt_solver import IPOPTSolver
from .optimization.optimization_problem import MinimizationProblem
from .optimization.optimization import (
    SciPyConvergenceError,
    minimize,
    maximize,
    print_optimization_methods,
)
from .control import Control
from .overloaded_type import OverloadedType, create_overloaded_object
from .verification import taylor_test, taylor_to_dict
from .drivers import compute_gradient, compute_derivative, compute_tlm, compute_hessian, solve_adjoint
from .checkpointing import disk_checkpointing_callback
from .reduced_functional import ReducedFunctional
from .adjfloat import AdjFloat, exp, log
from .tape import (
    Tape,
    set_working_tape,
    get_working_tape,
    no_annotations,
    annotate_tape,
    stop_annotating,
    pause_annotation,
    continue_annotation,
)
from .block import Block
from importlib.metadata import metadata

meta = metadata("pyadjoint-ad")
__version__ = meta["Version"]
__author__ = meta.get("Author", "")
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]


__all__ = [
    "Block",
    "Tape",
    "set_working_tape",
    "get_working_tape",
    "no_annotations",
    "annotate_tape",
    "stop_annotating",
    "pause_annotation",
    "continue_annotation",
    "AdjFloat",
    "exp",
    "log",
    "Control",
    "ReducedFunctional",
    "create_overloaded_object",
    "OverloadedType",
    "compute_gradient",
    "compute_derivative",
    "compute_tlm",
    "compute_hessian",
    "solve_adjoint",
    "taylor_test",
    "taylor_to_dict",
    "disk_checkpointing_callback",
    "MoolaOptimizationProblem",
    "InequalityConstraint",
    "EqualityConstraint",
    "TAOSolver",
    "ROLSolver",
    "IPOPTSolver",
    "MinimizationProblem",
    "SciPyConvergenceError",
    "minimize",
    "maximize",
    "print_optimization_methods",
]



from typing import Any
from .overloaded_type import OverloadedType, create_overloaded_object
import logging


class Control(object):
    """Defines a control variable from an OverloadedType.

    The control object references a specific node on the Tape. For mutable
    OverloadedType instances the Control only represents the value at the time
    of initialization.

    Example:
        Given a mutable OverloadedType instance u.

        >>> u = MutableFloat(1.0)
        >>> float(u)
        1.0
        >>> c1 = Control(u)
        >>> u.add_in_place(2.0)
        >>> c2 = Control(u)
        >>> float(u)
        3.0
        >>> c1.data()
        1.0
        >>> c2.data()
        3.0

        Now c1 represents the node prior to the add_in_place Block, while c2
        represents the node after the add_in_place Block. Creating a
        `ReducedFunctional` with c2 as Control results in a reduced problem
        without the add_in_place Block, while a ReducedFunctional with c1 as
        Control results in a forward model including the add_in_place.

    Args:
        control: The OverloadedType instance to define this control from.
        riesz_map: Parameters controlling how to find the Riesz representer of
            a dual (adjoint) variable to this control. The permitted values are
            type-dependent.

    """
    def __init__(self, control: OverloadedType, riesz_map: Any = None):
        self.control = control
        self.riesz_map = riesz_map
        self.block_variable = control.block_variable

    def data(self):
        return self.block_variable.checkpoint

    def tape_value(self):
        return create_overloaded_object(self.block_variable.saved_output)

    def get_derivative(self, apply_riesz=False):
        if self.block_variable.adj_value is None:
            logging.warning("Adjoint value is None, is the functional independent of the control variable?")
            return self.control._ad_init_zero(dual=not apply_riesz)
        elif apply_riesz:
            return self.control._ad_convert_riesz(
                self.block_variable.adj_value, riesz_map=self.riesz_map)
        else:
            return self.control._ad_init_object(self.block_variable.adj_value)

    def get_hessian(self, apply_riesz=False):
        if self.block_variable.hessian_value is None:
            logging.warning("Hessian value is None, is the functional independent of the control variable?")
            return self.control._ad_init_zero(dual=not apply_riesz)
        elif apply_riesz:
            return self.control._ad_convert_riesz(
                self.block_variable.hessian_value, riesz_map=self.riesz_map)
        else:
            return self.control._ad_init_object(
                self.block_variable.hessian_value
            )

    def update(self, value):
        # In the future we might want to call a static method
        # for converting a value to the correct type.
        # As this might depend on the OverloadedType control.
        if isinstance(value, OverloadedType):
            self.block_variable.checkpoint = value._ad_create_checkpoint()
        else:
            self.block_variable.checkpoint = value

    def update_numpy(self, value, offset):
        self.block_variable.checkpoint, offset =\
            self.assign_numpy(self.block_variable.checkpoint, value, offset)
        return offset

    def assign_numpy(self, dst, src, offset):
        # This returns dst and offset. dst needs to be returned in case
        # it is immutable. Because then we cannot assign in-place, but need
        # to create a new object.
        return self.control._ad_assign_numpy(dst, src, offset)

    def fetch_numpy(self, value):
        return self.control._ad_to_list(value)

    def copy_data(self):
        return self.control._ad_copy()

    @property
    def adj_value(self):
        return self.block_variable.adj_value

    @adj_value.setter
    def adj_value(self, value):
        self.block_variable.adj_value = value

    @property
    def tlm_value(self):
        return self.block_variable.tlm_value

    @tlm_value.setter
    def tlm_value(self, value):
        self.block_variable.tlm_value = value

    @property
    def hessian_value(self):
        return self.block_variable.hessian_value

    @hessian_value.setter
    def hessian_value(self, value):
        self.block_variable.hessian_value = value

    def __getattr__(self, item):
        return getattr(self.control, item)

    def mark_as_control(self):
        self.block_variable.is_control = True

    def unmark_as_control(self):
        self.block_variable.is_control = False



"""
Compatibility layer for ipopt wrappers
Currently only cyipopt is supported

Repos:

- https://github.com/matthias-k/cyipopt
"""

try:
    import cyipopt  # noqa: F401
except ImportError:
    raise ImportError("You need to install cyipopt. It is recommended to install IPOPT with HSL support!")



import numpy
from pyadjoint.overloaded_type import OverloadedType, register_overloaded_type, create_overloaded_object
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint.block import Block


@register_overloaded_type
class ndarray(OverloadedType, numpy.ndarray):
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def _ad_init_object(cls, obj):
        return cls(obj.shape, obj.dtype, buffer=obj)

    def _ad_create_checkpoint(self):
        return self.copy()

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def __getitem__(self, item):
        annotate = annotate_tape()
        if annotate:
            block = NumpyArraySliceBlock(self, item)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            out = numpy.ndarray.__getitem__(self, item)

        if annotate:
            out = create_overloaded_object(out)
            block.add_output(out.create_block_variable())
        return out

    def _ad_convert_type(self, value, options={}):
        return value

    def __array_finalize__(self, obj):
        OverloadedType.__init__(self)


class NumpyArraySliceBlock(Block):
    def __init__(self, array, item):
        super().__init__()
        self.add_dependency(array)
        self.item = item

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_output = numpy.zeros(inputs[0].shape)
        adj_output[self.item] = adj_inputs[0]
        return adj_output

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return inputs[0][self.item]



# flake8: noqa

import pyadjoint
__version__ = pyadjoint.__version__
__author__ = 'Sebastian Kenji Mitusch'
__credits__ = []
__license__ = 'LGPL-3'
__maintainer__ = 'Sebastian Kenji Mitusch'
__email__ = 'sebastkm@simula.no'


from .array import ndarray

# Use pyadjoint AdjFloat for numpy.float64.
import numpy
from pyadjoint.overloaded_type import register_overloaded_type
from pyadjoint.adjfloat import AdjFloat
register_overloaded_type(AdjFloat, numpy.float64)



try:
    from warnings import deprecated
except ImportError:
    from warnings import warn
    deprecated = None

from .enlisting import Enlist
from .tape import get_working_tape, stop_annotating


def compute_derivative(J, m, tape=None, adj_value=1.0, apply_riesz=False):
    """
    Compute the derivative of J with respect to the initialisation value of m,
    that is the value of m at its creation.

    Args:
        J (OverloadedType):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        tape: The tape to use. Default is the current tape.
        adj_value: The adjoint value to the result. Required if the functional
            is not scalar-valued, or if the functional is not the final stage
            in the computation of an outer functional.
        apply_riesz: If True, apply the Riesz map of each control in order
            to return a primal gradient rather than a derivative in the
            dual space.

    Returns:
        OverloadedType: The derivative with respect to the control.
            If apply_riesz is False, should be an instance of the type dual
            to that of the control. If apply_riesz is True should have the
            same type as the control.
    """
    tape = tape or get_working_tape()
    tape.reset_variables()
    J.block_variable.adj_value = adj_value
    m = Enlist(m)

    with stop_annotating():
        with tape.marked_control_dependents(m):
            with tape.marked_functional_dependencies(J):
                with marked_controls(m):
                    tape.evaluate_adj(markings=True)

        grads = [i.get_derivative(apply_riesz=apply_riesz) for i in m]
    return m.delist(grads)


def compute_gradient(J, m, tape=None, adj_value=1.0, apply_riesz=True):
    """
    Compute the gradient of J with respect to the initialisation value of m,
    that is the value of m at its creation.

    This function is deprecated in favour of :compute_derivative

    Args:
        J (OverloadedType):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        tape: The tape to use. Default is the current tape.
        adj_value: The adjoint value to the result. Required if the functional
            is not scalar-valued, or if the functional is not the final stage
            in the computation of an outer functional.
        apply_riesz: If True, apply the Riesz map of each control in order
            to return a primal gradient rather than a derivative in the
            dual space.

    Returns:
        OverloadedType: The gradient with respect to the control.
            If apply_riesz is False, should be an instance of the type dual
            to that of the control. If apply_riesz is True should have the
            same type as the control.
    """
    if deprecated is None:
        warn("compute_gradient is deprecated in favour of compute_derivative.",
             FutureWarning)

    return compute_derivative(J, m, tape, adj_value, apply_riesz)


if deprecated is not None:
    compute_gradient = deprecated(
        "compute_gradient is deprecated in favour of compute_derivative."
    )(compute_gradient)


def compute_hessian(J, m, m_dot, hessian_input=None, tape=None, evaluate_tlm=True, apply_riesz=False):
    """
    Compute the Hessian of J in a direction m_dot at the current value of m

    Args:
        J (AdjFloat):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        m_dot (list or instance of the control type): The direction in which to
            compute the Hessian.
        hessian_input (OverloadedType): The value to start the Hessian accumulation
            from after the TLM calculation. Uses zero initialised value if None.
        tape: The tape to use. Default is the current tape.
        apply_riesz: If True, apply the Riesz map of each control in order
            to return the (primal) Riesz representer of the Hessian
            action.
        evaluate_tlm (bool): Whether or not to compute the forward (TLM) part of
            the Hessian calculation.  If False, assumes that the tape has already
            been populated with the required TLM values.

    Returns:
        OverloadedType: The action of the Hessian in the direction m_dot.
            If apply_riesz is False, should be an instance of the type dual
            to that of the control. If apply_riesz is true should have the
            same type as the control.
    """
    tape = tape or get_working_tape()

    # fill the relevant tlm values on the tape
    if evaluate_tlm:
        compute_tlm(J, m, m_dot, tape)

    tape.reset_hessian_values()

    if hessian_input is None:
        J.block_variable.hessian_value = (
            J.block_variable.output._ad_init_zero(dual=True))
    else:
        J.block_variable.hessian_value = (
            J.block_variable.output._ad_init_object(hessian_input))

    m = Enlist(m)
    with stop_annotating():
        with tape.marked_control_dependents(m):
            with tape.marked_functional_dependencies(J):
                tape.evaluate_hessian(markings=True)

        r = [v.get_hessian(apply_riesz=apply_riesz) for v in m]
    return m.delist(r)


def compute_tlm(J, m, m_dot, tape=None):
    """
    Compute the tangent linear model of J in a direction m_dot at the current value of m

    Args:
        J (OverloadedType):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        m_dot (list or instance of the control type): The direction in which to
            compute the tangent linear model.
        tape: The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The action of the tangent linear model with respect to the control
            in direction m_dot. Should be an instance of the same type as the functional.
    """
    tape = tape or get_working_tape()
    tape.reset_tlm_values()

    m = Enlist(m)
    m_dot = Enlist(m_dot)

    for mi, mdi in zip(m, m_dot):
        mi.tlm_value = mdi

    with stop_annotating():
        with tape.marked_control_dependents(m):
            tape.evaluate_tlm(markings=True)

    return J._ad_init_object(J.block_variable.tlm_value)


def solve_adjoint(J, tape=None, adj_value=1.0):
    """
    Solve the adjoint problem for a functional J.

    This traverses the entire tape backwards, unlike `compute_gradient` which only works out those
    parts of the adjoint necessary to compute the sensitivity with respect to the specified control.
    As a result sensitivities with respect to all intermediate states are accumulated in the
    `adj_value` attribute of the associated block-variables. The adjoint solution of each solution
    step is stored in the `adj_sol` attribute of the corresponding solve block.

    Args:
        J (AdjFloat):  The objective functional.
        tape: The tape to use. Default is the current tape.
    """
    tape = tape or get_working_tape()
    tape.reset_variables()
    J.adj_value = adj_value

    with stop_annotating():
        tape.evaluate_adj(markings=False)


class marked_controls:
    """A context manager for marking controls.

    Note:
        This is a context manager for marking whether the class:'BlockVariable' is
        a control. On exiting the context, the class:'BlockVariable' that were
        marked as controls are automatically unmarked.

    Args:
        controls (list): A list of :class:`Control` to mark within the context manager.
    """
    def __init__(self, controls):
        self.controls = controls

    def __enter__(self):
        for control in self.controls:
            control.mark_as_control()

    def __exit__(self, *args):
        for control in self.controls:
            control.unmark_as_control()



import collections

from .constraints import Constraint, canonicalise
from ..overloaded_type import OverloadedType, create_overloaded_object
from ..reduced_functional import AbstractReducedFunctional

__all__ = ['MinimizationProblem', 'MaximizationProblem']


class OptimizationProblem(object):
    """A class that encapsulates all the information required to formulate a
    reduced optimisation problem. Don't instantiate this: instantiate
    a MinimizationProblem or a MaximizationProblem."""

    def __init__(self, reduced_functional, bounds=None, constraints=None):

        bounds = self.enlist(bounds)
        self.__check_arguments(reduced_functional, bounds, constraints)

        #: reduced_functional: a dolfin_adjoint.AbstractReducedFunctional
        #: object that encapsulates a Function and Control
        self.reduced_functional = reduced_functional

        #: bounds: lower and upper bounds for the control (optional). None means
        #: unbounded. if not None, then it must be a list of the same length as
        #: the number controls for the reduced_functional. Each entry in the list
        #: must be a tuple (lb, ub), where ub and lb are floats, or objects
        #: of the same kind as the control.
        self.bounds = bounds

        #: constraints: general (possibly nonlinear) constraints on the controls.
        #: None means no constraints, otherwise a Constraint object or a list of
        #: Constraints.
        self.constraints = canonicalise(constraints)

    def __check_arguments(self, reduced_functional, bounds, constraints):

        if type(self) is OptimizationProblem:
            raise TypeError("Instantiate a MinimizationProblem or MaximizationProblem.")

        if not isinstance(reduced_functional, AbstractReducedFunctional):
            raise TypeError("reduced_functional should be an AbstractReducedFunctional")

        if bounds is not None:
            if len(bounds) != len(reduced_functional.controls):
                raise TypeError("bounds should be of length number of controls of the AbstractReducedFunctional")
            for (bound, control) in zip(bounds, reduced_functional.controls):
                if len(bound) != 2:
                    raise TypeError("Each bound should be a tuple of length 2 (lb, ub)")

                for b in bound:
                    b = create_overloaded_object(b, suppress_warning=True)
                    klass = control.tape_value().__class__
                    if not (isinstance(b, (int, float, type(None), klass))):
                        raise TypeError("This pair (lb, ub) should be None, a float, or a %s." % klass)

        if not ((constraints is None)
                or (isinstance(constraints, Constraint))
                or (isinstance(constraints, list))):
            raise TypeError("constraints should be None or a Constraint or a list of Constraints")

    def enlist(self, bounds):
        """Make bounds into canonical format: a list (ideally the same length as the number of controls)."""
        if bounds is None:
            return None

        if not isinstance(bounds, collections.abc.Iterable):
            raise TypeError("bounds must be iterable.")

        should_i_make_a_damn_list = True

        if len(bounds) != 2:
            should_i_make_a_damn_list = False

        if len(bounds) == 2:  # support 'bounds=(lb, ub)' as well as 'bounds=[(lb, ub)]'
            for bound in bounds:
                if isinstance(bound, collections.abc.Iterable) and not isinstance(bound, OverloadedType):
                    should_i_make_a_damn_list = False

        if should_i_make_a_damn_list:
            return [bounds]
        else:
            return bounds


class MinimizationProblem(OptimizationProblem):
    """A class that encapsulates all the information required to formulate a
    reduced minimization problem."""
    pass


class MaximizationProblem(OptimizationProblem):
    """A class that encapsulates all the information required to formulate a
    reduced maximization problem."""
    pass



import weakref
from .block_variable import BlockVariable
from .tape import get_working_tape

_overloaded_types = {}


def get_overloaded_class(backend_class):
    return _overloaded_types[backend_class]


def create_overloaded_object(obj, suppress_warning=False):
    """Creates an OverloadedType instance corresponding `obj`.

    If an OverloadedType corresponding to `obj` has not been registered
    through `register_overloaded_type`, a RuntimeWarning will be issued.

    Args:
        obj (object): The object to create an overloaded instance from.
        suppress_warning (bool, optional): When set to True,
            suppresses warning message when a suitable overloaded class is not found.
            Default False.

    Returns:
        OverloadedType

    """
    if isinstance(obj, OverloadedType):
        return obj

    obj_type = type(obj)
    if obj_type in _overloaded_types:
        overloaded_type = _overloaded_types[obj_type]
        return overloaded_type._ad_init_object(obj)
    else:
        if not suppress_warning:
            import warnings
            warnings.warn("Could not find overloaded class of type '{}'.".format(obj_type), stacklevel=2)
        return obj


def register_overloaded_type(overloaded_type, classes=None):
    """Register an overloaded type for use in `create_overloaded_object`

    Overloaded types used with this function should have implemented a classmethod `_ad_create_object`.
    For usage as a decorator, OverloadedType should be the first base of `overloaded_type`, and `classes`
    the second base.

    Args:
        overloaded_type (type): The OverloadedType subclass to register.
        classes (type, tuple, optional): The original class/classes that this OverloadedType subclass
        overloads.

    Returns:
        type: returns only `overloaded_type` such that it can be used as a decorator.

    """
    if isinstance(classes, (tuple, list)):
        for cl in classes:
            register_overloaded_type(overloaded_type, classes=cl)
    else:
        if classes is None:
            classes = overloaded_type.__bases__[1]
        _overloaded_types[classes] = overloaded_type
    return overloaded_type


class Weakref:
    """Weakref which is picklable if the referenced object is picklable or
    None.

    Args:
        obj (:obj:`object`): The object to hold a weak reference to. None
            indicates a reference to no object.
    """

    def __init__(self, obj=None):
        self._init(obj)

    def _init(self, obj):
        if obj is None:
            self._obj = lambda: None
        else:
            self._obj = weakref.ref(obj)

    def __call__(self):
        return self._obj()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_obj"] = self()
        return state

    def __setstate__(self, state):
        state = state.copy()
        obj = state.pop("_obj")
        self.__dict__.update(state)
        self._init(obj)


class OverloadedType(object):
    """Base class for OverloadedType types.

    The purpose of each OverloadedType is to extend a type such that
    it can be referenced by blocks as well as overload basic mathematical
    operations such as __mul__, __add__, where they are needed.

    """

    def __init__(self, *args, **kwargs):
        self.clear_block_variable()

    @classmethod
    def _ad_init_object(cls, obj):
        """This method will often need to be overridden.

        The method should implement a way to reconstruct a new overloaded instance
        from a (possibly) not-overloaded instance.

        Args:
            obj: An instance of the original type

        Returns:
            OverloadedType: An overloaded instance which is considered the same as `obj`.

        """
        return cls(obj)

    def _ad_init_zero(self, dual=False):
        """This method must be overridden.

        Return a new overloaded zero of the appropriate type.

        If `dual` is `True`, return a zero of the dual type to this type. If
        the type is self-dual, this parameter is ignored. Note that by
        linearity there is no need to provide a riesz map in this case.

        Args:
            dual: Whether to return a primal or dual zero.

        Returns:
            OverloadedType: An object of the relevant type with the value zero.

        """
        raise NotImplementedError

    @property
    def block_variable(self):
        block_variable = self._block_variable()
        return self.create_block_variable() if block_variable is None else block_variable

    @block_variable.setter
    def block_variable(self, value):
        self._block_variable = Weakref(value)

    def clear_block_variable(self):
        self._block_variable = Weakref()

    def create_block_variable(self):
        self.block_variable = block_variable = BlockVariable(self)
        return block_variable

    def _ad_convert_riesz(self, value, riesz_map=None):
        """Apply a Riesz map to convert an adjoint result to a primal variable.

        Args:
            value (Any): The value to convert. Should be a result of an adjoint
                computation.
            riesz_map: Parameters controlling how to find the Riesz
                representer. The permitted values are type-dependent.

        Returns:
            OverloadedType: An instance of the same type as `self`.

        """
        raise NotImplementedError(f"OverloadedType._ad_convert_riesz not defined for class {type(self)}.")

    def _ad_create_checkpoint(self):
        """This method must be overridden.

        Should implement a way to create a checkpoint for the overloaded object.
        The checkpoint should be returned and possible to restore from in the
        corresponding _ad_restore_at_checkpoint method.

        Returns:
            :obj:`object`: A checkpoint. Could be of any type, but must be possible
                to restore an object from that point.

        """
        raise NotImplementedError

    def _ad_restore_at_checkpoint(self, checkpoint):
        """This method must be overridden.

        Should implement a way to restore the object at supplied checkpoint.
        The checkpoint is created from the _ad_create_checkpoint method.

        Returns:
            :obj:`OverloadedType`: The object with same state as at the supplied checkpoint.

        """
        raise NotImplementedError

    def _ad_mul(self, other):
        """This method must be overridden.

        The method should implement a routine for multiplying the overloaded object
        with another object, and return an object of the same type as `self`.

        Args:
            other (:obj:`object`): The object to be multiplied with this.
                Should at the very least accept :obj:`float` and :obj:`integer` objects.

        Returns:
            :obj:`OverloadedType`: The product of the two objects represented as
                an instance of the same subclass of :class:`OverloadedType` as the type
                of `self`.

        """
        raise NotImplementedError

    def _ad_imul(self, other):
        """In-place multiplies `self` with `other`.

        This method should be overridden if the default behaviour is not compatible with this OverloadedType.

        Args:
            other (object): The object to multiply `self` with.
                Should at the very least accept `float` objects.

        Returns:
            None

        """
        self *= other

    def _ad_add(self, other):
        """This method must be overridden.

        The method should implement a routine for adding the overloaded object
        with another object, and return an object of the same type as `self`.

        Args:
            other (:obj:`object`): The object to be added with this.
                Should at the very least accept objects of the same type as `self`.

        Returns:
            :obj:`OverloadedType`: The sum of the two objects represented as
                an instance of the same subclass of :class:`OverloadedType` as the type
                of `self`.

        """
        raise NotImplementedError

    def _ad_iadd(self, other):
        """In-place adds `other` to `self`.

        This method should be overridden if the default behaviour is not compatible with this OverloadedType.

        Args:
            other (object): The object to multiply `self` with.
                Should at the very least accept objects of the same type as `self`.

        Returns:
            None

        """
        self += other

    def _ad_dot(self, other):
        """This method must be overridden.

        The method should implement a routine for computing the dot product of
        the overloaded object with another object of the same type, and return
        a :obj:`float`.

        Args:
            other (:obj:`OverloadedType`): The object to compute the dot product with.
                Should be of the same type as `self`.

        Returns:
            :obj:`float`: The dot product of the two objects.

        """
        raise NotImplementedError

    def _ad_will_add_as_dependency(self):
        """Method called when the object is added as a Block dependency.

        """
        self.block_variable.save_output(overwrite=False)

    def _ad_will_add_as_output(self):
        """Method called when the object is added as a Block output.

        Returns:
            bool: True if the saved checkpoint should be overwritten.

        """
        return True

    def _ad_to_petsc(self, vec=None):
        """This method must be overridden.

        If the `vec` parameter is not provided, this method returns a new
        instance that is a copy of the PETSc Vec associated with the
        OverloadedType object. If the `vec` parameter is supplied, the PETSc
        Vec associated with the OverloadedType object is copied into the
        provided `vec`.

        Args:
            vec (PETSc.Vec, optional): A PETSc Vec to copy the data into.

        Returns:
            PETSc.Vec: A new instance that is a copy of the PETSc Vec of the
                overloaded object.
        """
        raise NotImplementedError

    def _ad_from_petsc(self, vec):
        """This method must be overridden.

        The method should implement a routine to update the PETSc Vec associated
        with the overloaded object from the given PETSc Vec `vec`.

        Args:
            vec (PETSc.Vec): The PETSc Vec from which to update the `self`
                associated PETSc Vec.

        """
        raise NotImplementedError

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        """This method must be overridden.

        The method should implement a routine for assigning the values from
        a numpy array `src` to the checkpoint `dst`. `dst` should be an instance
        of the implementing class.

        Args:
            dst (obj): The object which should be assigned new values.
                The type will most likely be an OverloadedType or similar.
            src (numpy.ndarray): The numpy array to use as a source for the assignment.
                `src` should have the same underlying dimensions as `dst`.
            offset (int): Start reading `dst` from `offset`.

        Returns:
            tuple:

                obj: The `dst` object. If `dst` is mutable it is preferred to be the same
                    instance as supplied to the function call. Otherwise a new instance
                    must be initialized and returned with the correct `src` values.

                int: The new offset.

        """
        raise NotImplementedError

    @staticmethod
    def _ad_to_list(m):
        """This method must be overridden.

        The method should implement a routine for converting `m` into a
        list type. `m` should be an instance of the same type as the class
        this method is implemented in. Although maybe the backend version
        of this class, meaning it is not necessarily an OverloadedType.

        Args:
            m (obj): The object to be converted into a list.

        Returns:
            list: A list representation of the data structure of `m`.

        """
        raise NotImplementedError

    def _ad_copy(self):
        """This method must be overridden.

        The method should implement a routine for copying itself.

        Returns:
            OverloadedType: A (deep) copy of `self`.

        """
        raise NotImplementedError

    def _ad_dim(self):
        """This method must be overridden.

        The method should implement a routine for computing the number of components
        of `self`.

        Returns:
            int: The number of components of `self`.

        """
        raise NotImplementedError

    @property
    def _ad_str(self):
        """Return the string representation of the block variable.

        For simple scalar types this might be the string representation of the
        stored value. For more complex objects it is likely to be based on a
        name attached to this instance.

        Returns:
            str: A human readable serialisation of this variable.
        """
        return str(self)


class FloatingType(OverloadedType):
    def __init__(self, *args, **kwargs):
        self.block_class = kwargs.pop("block_class", None)
        self._ad_args = kwargs.pop("_ad_args", [])
        self._ad_kwargs = kwargs.pop("_ad_kwargs", {})
        self.ad_block_tag = kwargs.pop("ad_block_tag", None)
        self._ad_floating_active = kwargs.pop("_ad_floating_active", False)
        self.block = None

        self._ad_output_args = kwargs.pop("_ad_output_args", [])
        self._ad_output_kwargs = kwargs.pop("_ad_output_kwargs", {})
        self.output_block_class = kwargs.pop("output_block_class", None)
        self._ad_outputs = kwargs.pop("_ad_outputs", [])
        OverloadedType.__init__(self, *args, **kwargs)

    def create_block_variable(self):
        block_variable = OverloadedType.create_block_variable(self)
        block_variable.floating_type = True
        return block_variable

    def _ad_will_add_as_dependency(self):
        if self._ad_floating_active:
            with FloatingType.stop_floating(self):
                self._ad_annotate_block()
        self.block_variable.save_output(overwrite=False)

    def _ad_will_add_as_output(self):
        if self._ad_floating_active:
            with FloatingType.stop_floating(self):
                self._ad_annotate_output_block()
        return True

    def _ad_annotate_block(self):
        if self.block_class is None:
            return

        tape = get_working_tape()
        block = self.block_class(*self._ad_args, **self._ad_kwargs)
        block.tag = self.ad_block_tag
        self.block = block
        tape.add_block(block)
        block.add_output(self.create_block_variable())

    def _ad_annotate_output_block(self):
        if self.output_block_class is None:
            return

        tape = get_working_tape()
        block = self.output_block_class(self, *self._ad_output_args, **self._ad_output_kwargs)
        self.output_block = block
        tape.add_block(block)
        for output in self._ad_outputs:
            block.add_output(output.create_block_variable())

    class stop_floating(object):
        def __init__(self, obj):
            self.obj = obj

        def __enter__(self):
            self.obj._ad_floating_active = False

        def __exit__(self, *args):
            self.obj._ad_floating_active = True



class Enlist(list):
    def __init__(self, x):
        self.listed = isinstance(x, (list, tuple))
        super(Enlist, self).__init__(x if self.listed else [x])

    def delist(self, y=None):
        y = self if y is None else y
        if self.listed:
            return y
        else:
            assert len(y) == 1
            return y[0]



from .tape import no_annotations
from html import escape


class Block(object):
    """Base class for all Tape Block types.

    Each instance of a Block type represents an elementary operation in the
    forward model.

    Abstract methods
        :func:`evaluate_adj`

    """
    __slots__ = ['_dependencies', '_outputs', 'block_helper', 'adj_state', 'tag']
    pop_kwargs_keys = []

    def __init__(self, ad_block_tag=None):
        self._dependencies = []
        self._outputs = []
        self.block_helper = None
        self.tag = ad_block_tag
        # The adjoint state of the block. This is used to store the adjoint
        # state, for instance to store the adjoint state of a solver.
        self.adj_state = None

    @classmethod
    def pop_kwargs(cls, kwargs):
        """Takes in a dictionary of keyword arguments,
        and pops the ones used by the Block-subclass `cls`
        """
        keys = cls.pop_kwargs_keys
        d = {}
        for k in keys:
            if k in kwargs:
                d[k] = kwargs.pop(k)
        return d

    def reset(self):
        if self.block_helper is not None:
            self.block_helper.reset()

    def add_dependency(self, dep, no_duplicates=False):
        """Adds object to the block dependencies.

        Will also save the output if it has not been saved before. Which should only happen if the
        BlockVariable was not created by a Block (but by the user).

        Args:
            dep (OverloadedType): The object to be added.
            no_duplicates (bool, optional): If True, the dependency is only added if it is not already in the list.
                Default is False.

        """
        if not no_duplicates or dep.block_variable not in self._dependencies:
            dep.block_variable.will_add_as_dependency()
            self._dependencies.append(dep.block_variable)

    def get_dependencies(self):
        """Returns the list of dependencies.

        Returns:
            :obj:`list`: A list of :class:`BlockVariable` instances.

        """
        return self._dependencies

    def add_output(self, obj):
        """Adds object to the block output list.

        Will also save the output.

        Args:
            obj (:class:`BlockVariable`): The object to be added.

        """
        obj.will_add_as_output()
        self._outputs.append(obj)

    def get_outputs(self):
        """Returns the list of block outputs.

        Returns:
            :obj:`list`: A list of :class:`BlockVariable` instances.

        """
        return self._outputs

    def reset_variables(self, types=None):
        """Resets all adjoint variables in the block dependencies and outputs.

        """
        types = ("adjoint",) if types is None else types

        for dep in self._dependencies:
            dep.reset_variables(types)

        for output in self._outputs:
            output.reset_variables(types)

    def reset_adjoint_state(self):
        """Resets the adjoint state of the block.
        """
        self.adj_state = None

    @no_annotations
    def evaluate_adj(self, markings=False):
        """Computes the adjoint action and stores the result in the `adj_value` attribute of the dependencies.

        This method will by default call the `evaluate_adj_component` method for each dependency.

        Args:
            markings (bool): If True, then each block_variable will have set `is_control_dependent` attribute
                indicating whether their adjoint components are relevant for computing the final target adjoint
                values and `is_functional_dependency` indicating whether its value impacts the functional value.
                Default is False.

        """
        outputs = self.get_outputs()
        adj_inputs = []
        has_input = False
        for output in outputs:
            adj_inputs.append(output.adj_value)
            if output.adj_value is not None:
                has_input = True

        if not has_input:
            return

        deps = self.get_dependencies()
        inputs = [bv.saved_output for bv in deps]
        relevant_dependencies = [(i, bv) for i, bv in enumerate(deps) if bv.is_control_dependent or not markings]

        if len(relevant_dependencies) <= 0:
            return

        prepared = self.prepare_evaluate_adj(inputs, adj_inputs, relevant_dependencies)

        for idx, dep in relevant_dependencies:
            adj_output = self.evaluate_adj_component(inputs,
                                                     adj_inputs,
                                                     dep,
                                                     idx,
                                                     prepared)
            if adj_output is not None:
                dep.add_adj_output(adj_output)

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        """Runs preparations before `evalute_adj_component` is ran.

        The return value is supplied to each of the subsequent `evaluate_adj_component` calls.
        This method is intended to be overridden for blocks that require such preparations, by default there is none.

        Args:
            inputs: The values of the inputs
            adj_inputs: The adjoint inputs
            relevant_dependencies: A list of the relevant block variables for `evaluate_adj_component`.

        Returns:
            Anything. The returned value is supplied to `evaluate_adj_component`

        """
        return None

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        """This method should be overridden.

        The method should implement a routine for evaluating the adjoint of the block that corresponds to
        one dependency.
        If one considers the adjoint action a vector right multiplied with the Jacobian matrix,
        then this method should return one entry in the resulting product, where the entry
        returned is decided by the argument `idx`.

        Args:
            inputs (list): A list of the saved input values, determined by the dependencies list.
            adj_inputs (list): A list of the adjoint input values, determined by the outputs list.
            block_variable (BlockVariable): The block variable of the dependency corresponding to index `idx`.
            idx (int): The index of the component to compute.
            prepared (object): Anything returned by the prepare_evaluate_adj method. Default is None.

        Returns:
            An object of a type consistent with the adj_value type of `block_variable`: The resulting product.

        """
        raise NotImplementedError(type(self))

    @no_annotations
    def evaluate_tlm(self, markings=False):
        """Computes the tangent linear action and stores the result in the `tlm_value` attribute of the outputs.

        This method will by default call the `evaluate_tlm_component` method for each output.

        Args:
            markings (bool): If True, then each block_variable will have set `is_control_dependent` attribute
                indicating whether their tlm components are relevant for computing the final target tlm values.
                Default is False.

        """
        deps = self.get_dependencies()
        tlm_inputs = []
        has_input = False
        for dep in deps:
            tlm_inputs.append(dep.tlm_value)
            if dep.tlm_value is not None:
                has_input = True

        if not has_input:
            return

        outputs = self.get_outputs()
        inputs = [bv.saved_output for bv in deps]
        relevant_outputs = [(i, bv) for i, bv in enumerate(outputs) if bv.is_control_dependent or not markings]

        if len(relevant_outputs) <= 0:
            return

        prepared = self.prepare_evaluate_tlm(inputs, tlm_inputs, relevant_outputs)

        for idx, out in relevant_outputs:
            tlm_output = self.evaluate_tlm_component(inputs,
                                                     tlm_inputs,
                                                     out,
                                                     idx,
                                                     prepared)
            if tlm_output is not None:
                out.add_tlm_output(tlm_output)

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        """Runs preparations before `evalute_tlm_component` is ran.

        The return value is supplied to each of the subsequent `evaluate_tlm_component` calls.
        This method is intended to be overridden for blocks that require such preparations, by default there is none.

        Args:
            inputs: The values of the inputs
            tlm_inputs: The tlm inputs
            relevant_outputs: A list of the relevant block variables for `evaluate_tlm_component`.

        Returns:
            Anything. The returned value is supplied to `evaluate_tlm_component`

        """
        return None

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        """This method should be overridden.

        The method should implement a routine for computing the tangent linear model of the block that corresponds to
        one output.
        If one considers the tangent linear action as a Jacobian matrix multiplied with a vector,
        then this method should return one entry in the resulting product, where the entry returned
        is decided by the argument `idx`.

        Args:
            inputs (list): A list of the saved input values, determined by the dependencies list.
            tlm_inputs (list): A list of the tlm input values, determined by the dependencies list.
            block_variable (BlockVariable): The block variable of the output corresponding to index `idx`.
            idx (int): The index of the component to compute.
            prepared (object): Anything returned by the prepare_evaluate_tlm method. Default is None.

        Returns:
            An object of the same type as `block_variable.saved_output`: The resulting product.

        """
        raise NotImplementedError("evaluate_tlm_component is not implemented for Block-type: {}".format(type(self)))

    @no_annotations
    def evaluate_hessian(self, markings=False):
        outputs = self.get_outputs()
        hessian_inputs = []
        adj_inputs = []
        has_input = False
        for output in outputs:
            hessian_inputs.append(output.hessian_value)
            adj_inputs.append(output.adj_value)
            if output.hessian_value is not None:
                has_input = True

        if not has_input:
            return

        deps = self.get_dependencies()
        inputs = [bv.saved_output for bv in deps]
        relevant_dependencies = [(i, bv) for i, bv in enumerate(deps) if bv.is_control_dependent or not markings]

        if len(relevant_dependencies) <= 0:
            return

        prepared = self.prepare_evaluate_hessian(inputs, hessian_inputs, adj_inputs, relevant_dependencies)

        for idx, dep in relevant_dependencies:
            hessian_output = self.evaluate_hessian_component(inputs,
                                                             hessian_inputs,
                                                             adj_inputs,
                                                             dep,
                                                             idx,
                                                             relevant_dependencies,
                                                             prepared)
            if hessian_output is not None:
                dep.add_hessian_output(hessian_output)

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs, relevant_dependencies):
        """Runs preparations before `evalute_hessian_component` is ran for each relevant dependency.

        The return value is supplied to each of the subsequent `evaluate_hessian_component` calls.
        This method is intended to be overridden for blocks that require such preparations, by default there is none.

        Args:
            inputs: The values of the inputs
            hessian_inputs: The hessian inputs
            adj_inputs: The adjoint inputs
            relevant_dependencies: A list of the relevant block variables for `evaluate_hessian_component`.

        Returns:
            Anything. The returned value is supplied to `evaluate_hessian_component`

        """
        return None

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        """This method must be overridden.

        The method should implement a routine for evaluating the hessian of the block.
        It is preferable that a "Forward-over-Reverse" scheme is used. Thus the hessians
        are evaluated in reverse (starting with the last block on the tape).

        """
        raise NotImplementedError(type(self))

    def recompute(self, markings=False):
        """Recomputes the overloaded function with new inputs
            and stores the results in the `checkpoint` attribute of the outputs.

        This method will by default call the `recompute_component` method for each output.

        Args:
            markings (bool): If True, then each block_variable will have set `is_control_dependent` attribute
                indicating whether their checkpoints need to be recomputed for recomputing the final target
                function value.
                Default is False.

        """
        outputs = self.get_outputs()
        for out in outputs:
            if out.is_control:
                # We assume that a Block with multiple outputs where one of them is a control
                # can not depend on any other controls.
                return

        inputs = [bv.saved_output for bv in self.get_dependencies()]
        relevant_outputs = [(i, bv) for i, bv in enumerate(outputs) if bv.is_control_dependent or not markings]

        if len(relevant_outputs) <= 0:
            return

        prepared = self.prepare_recompute_component(inputs, relevant_outputs)

        for idx, out in relevant_outputs:
            output = self.recompute_component(inputs,
                                              out,
                                              idx,
                                              prepared)
            if output is not None:
                out.checkpoint = output

    def prepare_recompute_component(self, inputs, relevant_outputs):
        """Runs preparations before `recompute_component` is ran.

        The return value is supplied to each of the subsequent `recompute_component` calls.
        This method is intended to be overridden for blocks that require such preparations, by default there is none.

        Args:
            inputs: The values of the inputs
            relevant_outputs: A list of the relevant block variables for `recompute_component`.

        Returns:
            Anything. The returned value is supplied to `recompute_component`

        """
        return None

    def recompute_component(self, inputs, block_variable, idx, prepared):
        """This method must be overridden.

        The method should implement a routine for recomputing one output of the block in the forward computations.
        The output to recompute is determined by the `idx` argument, which corresponds to the index
        of the output in the outputs list.
        If the block only has a single output, then `idx` will always be 0.

        Args:
            inputs (list): A list of the saved input values, determined by the dependencies list.
            block_variable (BlockVariable): The block variable of the output corresponding to index `idx`.
            idx (int): The index of the output to compute.
            prepared (object): Anything returned by the prepare_recompute_component method. Default is None.

        Returns:
            An object of the same type as `block_variable.checkpoint` which is determined by
            `OverloadedType._ad_create_checkpoint` (often the same as `block_variable.saved_output`): The new output.

        """
        raise NotImplementedError

    def create_graph(self, G, pos):
        # Edges for block dependencies
        for xpos, dep in enumerate(self.get_dependencies()):
            G.add_edge(id(dep), id(self))
            if "label" not in G.nodes[id(dep)]:
                G.nodes[id(dep)]['label'] = escape(str(dep))
                G.nodes[id(dep)]['node_color'] = "r"
                G.nodes[id(dep)]['position'] = (0.1 * xpos, -pos + 0.5)

        # Edges for block outputs
        for xpos, out in enumerate(self.get_outputs()):
            G.add_edge(id(self), id(out))
            if "label" not in G.nodes[id(out)]:
                G.nodes[id(out)]['label'] = escape(str(out))
                G.nodes[id(out)]['node_color'] = "r"
                G.nodes[id(out)]['position'] = (0.1 * xpos, -pos - 0.5)

        # Set properties for Block node
        G.nodes[id(self)]['label'] = escape(str(self))
        G.nodes[id(self)]['node_color'] = "b"
        G.nodes[id(self)]['position'] = (0, -pos)
        G.nodes[id(self)]['shape'] = "box"
        G.nodes[id(self)]['style'] = "filled"
        G.nodes[id(self)]['fillcolor'] = "lightgrey"



from enum import Enum
from numbers import Complex

import numpy as np

from ..enlisting import Enlist
from ..overloaded_type import OverloadedType
from .optimization_problem import MinimizationProblem
from .optimization_solver import OptimizationSolver


try:
    import petsc4py.PETSc as PETSc
except ModuleNotFoundError:
    PETSc = None
try:
    import petsctools
except ModuleNotFoundError:
    petsctools = None

__all__ = [
    "TAOConvergenceError",
    "TAOSolver"
]


class PETScVecInterface:
    """Interface for conversion between :class:`OverloadedType` objects and
    :class:`petsc4py.PETSc.Vec` objects.

    This uses the generic interface provided by :class:`OverloadedType`.

    Args:
        x (OverloadedType or Sequence[OverloadedType]): Defines the data
            layout.
        comm (petsc4py.PETSc.Comm or mpi4py.MPI.Comm): Communicator.
    """

    def __init__(self, x, *, comm=None):
        if PETSc is None:
            raise RuntimeError("PETSc not available")
        if petsctools is None:
            raise RuntimeError("petsctools not available")

        x = Enlist(x)
        comm = valid_comm(comm)

        vecs = tuple(x_i._ad_to_petsc() for x_i in x)
        n = sum(vec.getLocalSize() for vec in vecs)
        N = sum(vec.getSize() for vec in vecs)
        _, isets = PETSc.Vec().concatenate(vecs)

        self._comm = comm
        self._n = n
        self._N = N
        self._isets = isets

    @property
    def comm(self):
        """Communicator.
        """

        return self._comm

    @property
    def n(self):
        """Total number of process local DoFs."""

        return self._n

    @property
    def N(self):
        """Total number of global DoFs."""

        return self._N

    def new_petsc(self):
        """Construct a new :class:`petsc4py.PETSc.Vec`.

        Returns:
            petsc4py.PETSc.Vec: The new :class:`petsc4py.PETSc.Vec`.
        """

        vec = PETSc.Vec().create(comm=self.comm)
        vec.setSizes((self.n, self.N))
        vec.setUp()
        return vec

    def from_petsc(self, y, x):
        """Copy data from a :class:`petsc4py.PETSc.Vec` to variables.

        Args:
            y (petsc4py.PETSc.Vec): The input :class:`petsc4py.PETSc.Vec`.
            x (OverloadedType or Sequence[OverloadedType]): The output
                variables.
        """

        x = Enlist(x)
        if len(x) != len(self._isets):
            raise ValueError("Invalid length")
        for iset, x_i in zip(self._isets, x):
            y_sub = y.getSubVector(iset)
            x_i._ad_from_petsc(y_sub)
            y.restoreSubVector(iset, y_sub)

    def to_petsc(self, x, y):
        """Copy data to a :class:`petsc4py.PETSc.Vec`.

        Args:
            x (petsc4py.PETSc.Vec): The output :class:`petsc4py.PETSc.Vec`.
            y (Complex, OverloadedType or Sequence[OverloadedType]):
                Values for input variables.
        """

        y = Enlist(y)
        if len(y) != len(self._isets):
            raise ValueError("Invalid length")
        for iset, y_i in zip(self._isets, y):
            x_sub = x.getSubVector(iset)
            if isinstance(y_i, Complex):
                x_sub.set(y_i)
            elif isinstance(y_i, OverloadedType):
                y_i._ad_to_petsc(vec=x_sub)
            else:
                raise TypeError(f"Unexpected type: {type(y_i)}")
            x.restoreSubVector(iset, x_sub)


def new_control_variable(reduced_functional, *, dual=False):
    """Return new variables suitable for storing a control value or its dual.

    Args:
        reduced_functional (ReducedFunctional): The reduced functional whose
        controls are to be copied.
        dual (bool): whether to return a dual type. If False then a primal type is returned.

    Returns:
        tuple[OverloadedType]: New variables suitable for storing a control value.
    """
    return tuple(control._ad_init_zero(dual=dual)
                 for control in reduced_functional.controls)


def valid_comm(comm):
    """
    Return a valid communicator from a user provided Comm or None.

    Args:
        comm: Optional[Any[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]]

    Returns:
        petsc4py.PETSc.COMM_WORLD if `comm is None`, otherwise `comm.tompi4py()`.
    """
    if comm is None:
        comm = PETSc.COMM_WORLD
    if hasattr(comm, "tompi4py"):
        comm = comm.tompi4py()
    return comm


class RFOperation(Enum):
    """
    The type of linear action that a ReducedFunctionalMat should apply.
    """
    FORWARD = 'forward'
    TLM = 'tlm'
    ADJOINT = 'adjoint'
    HESSIAN = 'hessian'


class ReducedFunctionalMatBase:
    """
    PETSc.Mat Python context to apply the action of a pyadjoint.ReducedFunctional.

    If V is the control space and U is the functional space, each action has the following map:
    Jhat : V -> U
    TLM : V -> U
    Adjoint : U* -> V*
    Hessian : V x U* -> V* | V -> V*

    Child classes must implement:
    - mult_impl
    - multHermitian_impl
    - update_adjoint

    Args:
        rf (ReducedFunctional): Defines the forward model. Used to compute Mat actions.
        apply_riesz (bool): Whether to apply the riesz map before returning the
            result of the action to PETSc.
        appctx (Optional[dict]): User provided context.
        always_update_tape (bool): Whether to force reevaluation of the forward model
            every time `mult` is called. If needs_adjoint_update then this will also force
            the adjoint model to be reevaluated at every call to `mult`.
        needs_functional_interface: Whether to create a PETScVecInterface for rf.functional.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the rf is defined over.
    """

    def __init__(self, rf, *,
                 apply_riesz=False, appctx=None,
                 always_update_tape=False,
                 needs_functional_interface=False,
                 comm=None):
        comm = valid_comm(comm)

        self.rf = rf
        self.appctx = appctx
        self.apply_riesz = apply_riesz

        self.control_interface = PETScVecInterface(
            tuple(c.control for c in rf.controls),
            comm=comm)
        if needs_functional_interface:
            self.functional_interface = PETScVecInterface(
                rf.functional, comm=comm)

        self._m = new_control_variable(rf)
        self._shift = 0
        self.always_update_tape = always_update_tape

    @classmethod
    def update(cls, obj, x, A, P):
        ctx = A.getPythonContext()
        ctx.control_interface.from_petsc(x, ctx._m)
        ctx.update_tape_values(update_adjoint=cls.update_adjoint())
        ctx._shift = 0

    def shift(self, A, alpha):
        self._shift += alpha

    def update_tape_values(self, *, update_adjoint=True):
        _ = self.rf(self._m)
        if update_adjoint:
            _ = self.rf.derivative(apply_riesz=False)

    def mult(self, A, x, y):
        self.xinterface.from_petsc(x, self.x)
        out = self.mult_impl(A, self.x)
        self.yinterface.to_petsc(y, out)

        if self._shift != 0:
            y.axpy(self._shift, x)

    def multHermitian(self, A, x, y):
        self.yinterface.from_petsc(x, self.x)
        out = self.multHermitian_impl(A, self.x)
        self.xinterface.to_petsc(y, out)

        if self._shift != 0:
            y.axpy(self._shift, x)

    def mult_impl(self, A, x):
        """
        This method must be overriden.

        Provides the implementation of a particular type of ReducedFunctional action.

        Args:
            A (PETSc.Mat): The Mat that this python context is attached to.
            x (Union[OverloadedType, list[OverloadedType]]): An element in either the control
                or functional space of the ReducedFunctional that this Mat will act on.
        """
        raise NotImplementedError(
            "Must provide implementation of the action of this matrix on an OverloadedType")

    def multHermitian_impl(self, A, y):
        """
        This method must be overriden.

        Provides the implementation of the Hermitian of a particular type of ReducedFunctional action.

        Args:
            A (PETSc.Mat): The Mat that this python context is attached to.
            y (Union[OverloadedType, list[OverloadedType]]): An element in either the control
                or functional space of the ReducedFunctional that this Mat will act on.
        """
        raise NotImplementedError(
            "Must provide implementation of the Hermitian action of this matrix on an OverloadedType")

    @classmethod
    def update_adjoint(self):
        """
        Whether to update the adjoint as well as the forward model when updating
        the Mat. If True then `rf.derivative` will be called by the `update` method.
        """
        raise NotImplementedError


class ReducedFunctionalHessianMat(ReducedFunctionalMatBase):
    """
    PETSc.Mat Python context to apply the Hessian action of a pyadjoint.ReducedFunctional.

    If V is the control space and U is the functional space, the Hessian action has the following map:
    Jhat : V -> U
    Hessian : V x U* -> V* | V -> V*

    Args:
        rf (ReducedFunctional): Defines the forward model, and used to compute operator actions.
        apply_riesz (bool): Whether to apply the riesz map before returning the result of the
            action to PETSc.
        appctx (Optional[dict]): User provided context.
        always_update_tape (bool): Whether to force reevaluation of the forward and adjoint models
            every time `mult` is called.
        comm (Optional[Union[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]]): Communicator that the rf is defined over.
    """

    def __init__(self, rf, *, apply_riesz=False, appctx=None,
                 always_update_tape=False, comm=None):

        super().__init__(rf, apply_riesz=apply_riesz, appctx=appctx,
                         needs_functional_interface=False,
                         always_update_tape=always_update_tape, comm=comm)

        self.xinterface = self.control_interface
        self.yinterface = self.control_interface
        self.x = new_control_variable(rf)

    @classmethod
    def update_adjoint(self):
        return True

    def mult_impl(self, A, x):
        if self.always_update_tape:
            self.update_tape_values(update_adjoint=True)
        return self.rf.hessian(
            x, apply_riesz=self.apply_riesz)

    def multHermitian_impl(self, A, x):
        return self.mult_impl(A, x)


class ReducedFunctionalAdjointMat(ReducedFunctionalMatBase):
    """
    PETSc.Mat Python context to apply the adjoint action of a pyadjoint.ReducedFunctional.

    If V is the control space and U is the functional space, the adjoint action has the following map:
    Jhat : V -> U
    Adjoint : U* -> V*

    Args:
        rf (ReducedFunctional): Defines the forward model, and used to compute operator actions.
        apply_riesz (bool): Whether to apply the riesz map before returning the result of the
            action to PETSc.
        appctx (Optional[dict]): User provided context.
        always_update_tape (bool): Whether to force reevaluation of the forward model every time
            `mult` is called.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the rf is defined over.
    """

    def __init__(self, rf, *, apply_riesz=False, appctx=None,
                 always_update_tape=False, comm=None):

        super().__init__(rf, apply_riesz=apply_riesz, appctx=appctx,
                         needs_functional_interface=True,
                         always_update_tape=always_update_tape, comm=comm)

        self.xinterface = self.functional_interface
        self.yinterface = self.control_interface
        self.x = rf.functional._ad_copy()

    @classmethod
    def update_adjoint(self):
        return False

    def mult_impl(self, A, x):
        if self.always_update_tape:
            self.update_tape_values(update_adjoint=False)
        return self.rf.derivative(
            adj_input=x, apply_riesz=self.apply_riesz)

    def multHermitian_impl(self, A, x):
        if self.always_update_tape:
            self.update_tape_values(update_adjoint=False)
        return self.rf.tlm(x)


class ReducedFunctionalTLMMat(ReducedFunctionalMatBase):
    """
    PETSc.Mat Python context to apply the tangent linear action of a pyadjoint.ReducedFunctional.

    If V is the control space and U is the functional space, the tangent linear action has the following map:
    Jhat : V -> U
    TLM : V -> U

    Args:
        rf (ReducedFunctional): Defines the forward model, and used to compute operator actions.
        appctx (Optional[dict]): User provided context.
        always_update_tape (bool): Whether to force reevaluation of the forward model every time
            `mult` is called.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the rf is defined over.
    """

    def __init__(self, rf, *, appctx=None, always_update_tape=False, comm=None):
        super().__init__(rf, appctx=appctx, needs_functional_interface=True,
                         always_update_tape=always_update_tape, comm=comm)

        self.xinterface = self.control_interface
        self.yinterface = self.functional_interface
        self.x = new_control_variable(rf)

    @classmethod
    def update_adjoint(self):
        return False

    def mult_impl(self, A, x):
        if self.always_update_tape:
            self.update_tape_values(update_adjoint=False)
        return self.rf.tlm(x)

    def multHermitian_impl(self, A, x):
        if self.always_update_tape:
            self.update_tape_values(update_adjoint=False)
        return self.rf.derivative(
            adj_input=x, apply_riesz=self.apply_riesz)


def ReducedFunctionalMat(rf, action=RFOperation.HESSIAN, *, apply_riesz=False, appctx=None,
                         always_update_tape=False, comm=None):
    """
    PETSc.Mat to apply the action of a pyadjoint.ReducedFunctional.

    If V is the control space and U is the functional space, each action has the following map:
    Jhat : V -> U
    TLM : V -> U
    Adjoint : U* -> V*
    Hessian : V x U* -> V* | V -> V*

    Args:
        rf (ReducedFunctional): Defines the forward model, and used to compute operator actions.
        action (RFOperation): Whether to apply the TLM, adjoint, or Hessian action.
        apply_riesz (bool): Whether to apply the riesz map before returning the
            result of the action to PETSc. Ignored if action is TLM.
        appctx (Optional[dict]): User provided context.
        always_update_tape (bool): Whether to force reevaluation of the forward model every time
            `mult` is called. If action is HESSIAN then this will also force the adjoint model to
            be reevaluated at every call to `mult`.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the rf is defined over.
    """
    if action == RFOperation.HESSIAN:
        ctx = ReducedFunctionalHessianMat(
            rf, appctx=appctx, apply_riesz=apply_riesz,
            always_update_tape=always_update_tape, comm=comm)

    elif action == RFOperation.ADJOINT:
        ctx = ReducedFunctionalAdjointMat(
            rf, appctx=appctx, apply_riesz=apply_riesz,
            always_update_tape=always_update_tape, comm=comm)

    elif action == RFOperation.TLM:
        ctx = ReducedFunctionalTLMMat(
            rf, appctx=appctx,
            always_update_tape=always_update_tape, comm=comm)

    else:
        raise ValueError(
            f'Unrecognised RFOperation: {action}.')

    ncol = ctx.xinterface.n
    Ncol = ctx.xinterface.N

    nrow = ctx.yinterface.n
    Nrow = ctx.yinterface.N

    mat = PETSc.Mat().createPython(
        ((nrow, Nrow), (ncol, Ncol)),
        ctx, comm=ctx.control_interface.comm)
    if action == RFOperation.HESSIAN:
        mat.setOption(PETSc.Mat.Option.SYMMETRIC, True)
    mat.setUp()
    mat.assemble()
    return mat


class RieszMapMatCtx:
    """
    PETSc.Mat Python context to apply the Riesz map to an element in the dual of the control space.

    If V is the control space then this has the followiung signature:
    RieszMap : V* -> V

    Args:
        controls (Union[Control,list[Control]]): The controls defining the primal control space.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the controls are defined over.
    """
    def __init__(self, controls, comm=None):
        comm = valid_comm(comm)

        self.controls = Enlist(controls)
        self.vec_interface = PETScVecInterface(
            tuple(c.control for c in controls),
            comm=comm)

        self.dJ = tuple(c._ad_init_zero(dual=True)
                        for c in self.controls)

    def mult(self, mat, x, y):
        self.vec_interface.from_petsc(x, self.dJ)
        dJ = tuple(c._ad_convert_riesz(dJi, riesz_map=c.riesz_map)
                   for c, dJi in zip(self.controls, self.dJ))
        self.vec_interface.to_petsc(y, dJ)


def RieszMapMat(controls, symmetric=True, comm=None):
    """
    PETSc.Mat to apply the Riesz map to an element in the dual of the control space.

    If V is the control space then this has the followiung signature:
    RieszMap : V* -> V

    Args:
        controls (Union[Control,list[Control]]): The controls defining the primal control space.
        symmetric (bool): Whether the Riesz map attached to the Control is symmetric.
        comm (Optional[petsc4py.PETSc.Comm,mpi4py.MPI.Comm]): Communicator that the controls are defined over.
    """
    ctx = RieszMapMatCtx(controls, comm=comm)

    n = ctx.vec_interface.n
    N = ctx.vec_interface.N

    mat = PETSc.Mat().createPython(
        ((n, N), (n, N)), ctx,
        comm=ctx.vec_interface.comm)
    if symmetric:
        mat.setOption(PETSc.Mat.Option.SYMMETRIC, True)
    mat.setUp()
    mat.assemble()
    return mat


class TAOObjective:
    """Utility class for computing functional values and associated derivatives.

    Args:
        rf (AbstractReducedFunctional): Defines the forward, and used to
            compute derivative information.
        always_update_tape (bool): Whether to force reevaluation of the forward model every time
            gradient or hessian is called. If hessian is called then this will also force the
            adjoint model to be reevaluated.
    """

    def __init__(self, rf, *, always_update_tape=True):
        self._reduced_functional = rf
        self.always_update_tape = always_update_tape

    @property
    def reduced_functional(self):
        """:class:`.AbstractReducedFunctional`. Defines the forward, and used
        to compute derivative information.
        """
        return self._reduced_functional

    def objective(self, m):
        """Evaluate the forward.

        Args:
            m (OverloadedType or Sequence[OverloadedType]): Defines the control
                value.
        Returns:
            AdjFloat: The value of the functional.
        """
        m = Enlist(m)
        J = self.reduced_functional(tuple(m_i._ad_copy() for m_i in m))
        return J

    def gradient(self, m):
        """Compute a first derivative.

        Args:
            m (OverloadedType or Sequence[OverloadedType]): Defines the control
                value.
        Returns:
            AdjFloat: The value of the functional.
            OverloadedType or Sequence[OverloadedType]: The (dual space)
                derivative.
        """
        m = Enlist(m)
        if self.always_update_tape:
            _ = self.reduced_functional(tuple(m_i._ad_copy() for m_i in m))
        dJ = self.reduced_functional.derivative()
        return m.delist(dJ)

    def objective_gradient(self, m):
        """Evaluate the forward, and compute a first derivative.

        Args:
            m (OverloadedType or Sequence[OverloadedType]): Defines the control
                value.
        Returns:
            AdjFloat: The value of the functional.
            OverloadedType or Sequence[OverloadedType]: The (dual space)
                derivative.
        """

        m = Enlist(m)
        J = self.reduced_functional(tuple(m_i._ad_copy() for m_i in m))
        dJ = self.reduced_functional.derivative()
        return J, m.delist(dJ)

    def hessian(self, m, m_dot):
        """Evaluate the forward, and compute a second derivative action on a
        given direction.

        Args:
            m (OverloadedType or Sequence[OverloadedType]): Defines the control
                value.
            m_dot (OverloadedType or Sequence[OverloadedType]): Defines the
                direction.
        Returns:
            OverloadedType or Sequence[OverloadedType]: The (dual space)
                second derivative action on the given direction.
        """

        m = Enlist(m)
        m_dot = Enlist(m_dot)
        if self.always_update_tape:
            _ = self.reduced_functional(tuple(m_i._ad_copy() for m_i in m))
            _ = self.reduced_functional.derivative()
        ddJ = self.reduced_functional.hessian(tuple(m_dot_i._ad_copy() for m_dot_i in m_dot))
        return m.delist(ddJ)


class TAOConvergenceError(Exception):
    """Raised if a TAO solve fails to converge.
    """


if PETSc is None:
    _tao_reasons = {}
else:
    # Same approach as in _make_reasons in firedrake/solving_utils.py,
    # Firedrake master branch 57e21cc8ebdb044c1d8423b48f3dbf70975d5548
    _tao_reasons = {getattr(PETSc.TAO.Reason, key): key
                    for key in dir(PETSc.TAO.Reason)
                    if not key.startswith("_")}


class TAOSolver(OptimizationSolver):
    """Use TAO to solve an optimization problem.

    Args:
        problem (MinimizationProblem): Defines the optimization problem to be solved.
        parameters (Mapping): TAO options.
        options_prefix (Optional[str]): prefix for the TAO solver.
        appctx (Optional[dict]): User provided context.
        Pmat (Optional petsc4py.PETSc.Mat): Hessian preconditioning matrix.
        comm (petsc4py.PETSc.Comm or mpi4py.MPI.Comm): Communicator.
    """

    def __init__(self, problem, parameters, *,
                 options_prefix=None, appctx=None,
                 Pmat=None, comm=None):
        if PETSc is None:
            raise RuntimeError("PETSc not available")
        if petsctools is None:
            raise RuntimeError("petsctools not available")

        if not isinstance(problem, MinimizationProblem):
            raise TypeError("MinimizationProblem required")
        if problem.constraints is not None:
            raise NotImplementedError("Constraints not implemented")

        comm = valid_comm(comm)

        rf = problem.reduced_functional
        tao_objective = TAOObjective(rf)

        vec_interface = PETScVecInterface(
            tuple(control.control for control in rf.controls), comm=comm)

        tao = PETSc.TAO().create(comm=comm)

        def objective(tao, x):
            m = new_control_variable(rf)
            vec_interface.from_petsc(x, m)
            J_val = tao_objective.objective(m)
            return J_val

        def gradient(tao, x, g):
            m = new_control_variable(rf)
            vec_interface.from_petsc(x, m)
            dJ = tao_objective.gradient(m)
            vec_interface.to_petsc(g, dJ)

        def objective_gradient(tao, x, g):
            m = new_control_variable(rf)
            vec_interface.from_petsc(x, m)
            J_val, dJ = tao_objective.objective_gradient(m)
            vec_interface.to_petsc(g, dJ)
            return J_val

        tao.setObjectiveGradient(objective_gradient)
        tao.setObjective(objective)
        tao.setGradient(gradient)

        hessian_mat = ReducedFunctionalMat(
            problem.reduced_functional, appctx=appctx,
            action=RFOperation.HESSIAN, comm=comm)

        tao.setHessian(
            hessian_mat.getPythonContext().update,
            H=hessian_mat, P=Pmat or hessian_mat)

        Minv_mat = RieszMapMat(rf.controls, comm=comm)
        tao.setGradientNorm(Minv_mat)

        if problem.bounds is not None:
            lbs = []
            ubs = []
            assert len(problem.bounds) == len(problem.reduced_functional.controls)
            for (lb, ub), control in zip(problem.bounds, tao_objective.reduced_functional.controls):
                if lb is None:
                    lb = np.finfo(PETSc.ScalarType).min
                if ub is None:
                    ub = np.finfo(PETSc.ScalarType).max
                lbs.append(lb)
                ubs.append(ub)

            lb_vec = vec_interface.new_petsc()
            ub_vec = vec_interface.new_petsc()
            vec_interface.to_petsc(lb_vec, lbs)
            vec_interface.to_petsc(ub_vec, ubs)
            tao.setVariableBounds(lb_vec, ub_vec)

        petsctools.set_from_options(
            tao, parameters=parameters,
            options_prefix=options_prefix,
            default_prefix="pyadjoint")

        if tao.getType() in {PETSc.TAO.Type.LMVM, PETSc.TAO.Type.BLMVM}:
            n, N = vec_interface.n, vec_interface.N

            class InitialHessian:
                """:class:`petsc4py.PETSc.Mat` context.
                """

            class InitialHessianPreconditioner:
                """:class:`petsc4py.PETSc.PC` context.
                """
                def apply(self, pc, x, y):
                    Minv_mat.mult(x, y)

            # B_0_matrix is the initial Hessian approximation (following
            # Nocedal and Wright doi: 10.1007/978-0-387-40065-5 notation). This
            # is H0 in PETSc/TAO.

            B_0_matrix = PETSc.Mat().createPython(((n, N), (n, N)),
                                                  InitialHessian(), comm=comm)
            B_0_matrix.setOption(PETSc.Mat.Option.SYMMETRIC, True)
            B_0_matrix.setUp()

            B_0_matrix_pc = PETSc.PC().createPython(
                InitialHessianPreconditioner(), comm=comm)
            B_0_matrix_pc.setOperators(B_0_matrix)
            B_0_matrix_pc.setUp()

            tao.setLMVMH0(B_0_matrix)
            ksp = tao.getLMVMH0KSP()
            ksp.setType(PETSc.KSP.Type.PREONLY)
            ksp.setTolerances(rtol=0.0, atol=0.0, divtol=None, max_it=1)
            ksp.setPC(B_0_matrix_pc)
            ksp.setUp()
        else:
            B_0_matrix = None
            B_0_matrix_pc = None

        x = vec_interface.new_petsc()
        tao.setSolution(x)
        with petsctools.inserted_options(tao):
            tao.setUp()

        super().__init__(problem, parameters)
        self._tao_objective = tao_objective
        self._vec_interface = vec_interface
        self._tao = tao
        self._x = x

    @property
    def tao_objective(self):
        """The :class:`.TAOObjective` used for the optimization.
        """

        return self._tao_objective

    @property
    def tao(self):
        """The :class:`petsc4py.PETSc.TAO` used for the optimization.
        """

        return self._tao

    @property
    def x(self):
        """The :class:`petsc4py.PETSc.Vec` used to store the solution to the
        optimization problem.
        """

        return self._x

    def solve(self):
        """Solve the optimization problem.

        Returns:
            OverloadedType or Sequence[OverloadedType]: The solution.
        """

        controls = self.tao_objective.reduced_functional.controls
        m = tuple(control.tape_value()._ad_copy() for control in controls)

        with petsctools.inserted_options(self.tao):
            self._vec_interface.to_petsc(self.x, m)
            self.tao.solve()
            self._vec_interface.from_petsc(self.x, m)

        if self.tao.getConvergedReason() <= 0:
            # Using the same format as Firedrake linear solver errors
            raise TAOConvergenceError(
                f"TAOSolver failed to converge after {self.tao.getIterationNumber()} iterations "
                f"with reason: {_tao_reasons[self.tao.getConvergedReason()]}")
        if isinstance(controls, Enlist):
            return controls.delist(m)
        else:
            return m



"""
The optimization module is automatically imported by dolfin-adjoint
"""

__author__ = 'Simon Funke and Patrick Farrell'
__credits__ = ['Patrick Farrell', 'Simon Funke', 'David Ham', 'Marie Rognes']
__license__ = 'LGPL-3'
__maintainer__ = 'Simon Funke'
__email__ = 'simon@simula.no'



from .optimization_solver import OptimizationSolver
from ..enlisting import Enlist
from ..overloaded_type import OverloadedType
from ..tape import no_annotations

try:
    import ROL

    class ROLObjective(ROL.Objective):
        def __init__(self, rf, scale=1.):
            super(ROLObjective, self).__init__()
            self.rf = rf
            self.scale = scale

            self._val = None
            self._cache = None
            self._flag = None

        def value(self, x, tol):
            return self._val

        def gradient(self, g, x, tol):
            self.deriv = self.rf.derivative(apply_riesz=True)
            g.dat = Enlist(self.deriv)

        def hessVec(self, hv, v, x, tol):
            hessian_action = self.rf.hessian(v.dat, apply_riesz=True)
            hv.dat = Enlist(hessian_action)

        def update(self, x, flag, iteration):
            if hasattr(ROL, "UpdateType") and isinstance(flag, ROL.UpdateType):
                # Initial: has not been called before
                # Accept: this is the new iterate, trial has been called
                # Revert: revert to previous, trial has been called
                # Trial: candidate for next
                # Temp: temporary
                if flag in [ROL.UpdateType.Initial, ROL.UpdateType.Trial, ROL.UpdateType.Temp]:
                    self._val = self.rf(x.dat)
                    self._tape_trial = self.rf.tape.checkpoint_block_vars(self.rf.controls)
                elif flag == ROL.UpdateType.Revert:
                    # If we have cached value revert back to the cached value from _cache
                    if hasattr(self, "_tape_cache"):
                        self._val = self._cache
                        self.rf.tape.restore_block_vars(self._tape_cache)
                    # else (e.g., when we are unfortunate and restore from a reverted checkpoint)
                    # repopulate the tape for x.dat, and put them on _cache and _tape_cache
                    # as they were the last UpdateType.Accept values
                    else:
                        self._val = self.rf(x.dat)
                        self._cache = self._val
                        self._tape_cache = self._tape_trial

                self._flag = flag

                # cache value/tape in the first instance or when accepted
                if flag in [ROL.UpdateType.Initial, ROL.UpdateType.Accept]:
                    self._cache = self._val
                    self._tape_cache = self._tape_trial

            else:
                self._val = self.rf(x.dat)

    class ROLVector(ROL.Vector):
        def __init__(self, dat, inner_product="L2"):
            super(ROLVector, self).__init__()
            self.dat = dat
            self.inner_product = inner_product

        def plus(self, yy):
            for (x, y) in zip(self.dat, yy.dat):
                x._ad_iadd(y)

        def scale(self, alpha):
            for x in self.dat:
                x._ad_imul(alpha)

        def riesz_map(self, derivs):
            dat = []
            for deriv in Enlist(derivs):
                dat.append(deriv.riesz_representation(riesz_map=self.inner_product))
            return dat

        def dot(self, yy):
            res = 0.
            opts = {"riesz_representation": self.inner_product}
            for (x, y) in zip(self.dat, yy.dat):
                res += x._ad_dot(y, options=opts)
            return res

        def norm(self):
            return self.dot(self) ** 0.5

        def clone(self):
            dat = []
            for x in self.dat:
                dat.append(x._ad_copy())
            res = ROLVector(dat, inner_product=self.inner_product)
            res.scale(0.0)
            return res

        def dimension(self):
            return sum(x._ad_dim() for x in self.dat)

        def reduce(self, r, r0):
            res = r0
            for x in self.dat:
                res = x._reduce(r, res)
            return res

        def applyUnary(self, f):
            for x in self.dat:
                x._applyUnary(f)

        def applyBinary(self, f, inp):
            for (x, y) in zip(self.dat, inp.dat):
                x._applyBinary(f, y)

    class ROLConstraint(ROL.Constraint):

        def __init__(self, con):
            ROL.Constraint.__init__(self)
            self.con = con

        def value(self, cvec, x, tol):
            cvec.dat = self.con.function(x.dat)

        def applyJacobian(self, jv, v, x, tol):
            self.con.jacobian_action(x.dat, v.dat[0], jv.dat)

        def applyAdjointJacobian(self, jv, v, x, tol):
            self.con.jacobian_adjoint_action(x.dat, v.dat, jv.dat[0])
            jv.dat = jv.riesz_map(jv.dat)

        def applyAdjointHessian(self, ahuv, u, v, x, tol):
            self.con.hessian_action(x.dat, u.dat[0], v.dat, ahuv.dat[0])
            ahuv.dat = ahuv.riesz_map(ahuv.dat)

    class ROLSolver(OptimizationSolver):
        """
        Use ROL to solve the given optimisation problem.
        """

        def __init__(self, problem, parameters, inner_product="L2"):
            """
            Create a new ROLSolver.

            The argument inner_product specifies the inner product to be used for
            the control space.

            """

            OptimizationSolver.__init__(self, problem, parameters)
            self.rolobjective = ROLObjective(problem.reduced_functional)
            x = [p.tape_value() for p in self.problem.reduced_functional.controls]
            self.rolvector = ROLVector(x, inner_product=inner_product)
            self.params_dict = parameters

            self.bounds = self.__get_bounds()
            self.constraints = self.__get_constraints()

        def __get_bounds(self):
            bounds = self.problem.bounds
            if bounds is None:
                return None

            controlvec = self.rolvector
            lowervec = controlvec.clone()
            uppervec = controlvec.clone()

            for i in range(len(controlvec.dat)):
                general_lb, general_ub = bounds[i]
                if isinstance(general_lb, (int, float)):
                    lowervec.dat[i]._applyUnary(lambda x: general_lb)
                else:
                    lowervec.dat[i].assign(general_lb)
                if isinstance(general_ub, (int, float)):
                    uppervec.dat[i]._applyUnary(lambda x: general_ub)
                else:
                    uppervec.dat[i].assign(general_ub)

            res = ROL.Bounds(lowervec, uppervec, 1.0)
            # FIXME: without this the lowervec and uppervec get cleaned up too
            # early.  This is a bug in PyROL and we'll hopefully figure that out
            # soon
            self.lowervec = lowervec
            self.uppervec = uppervec
            return res

        def __get_constraints(self):
            if self.problem.constraints is None:
                return ([], []), ([], [])

            eqconstraints = self.problem.constraints.equality_constraints()

            if len(eqconstraints.constraints) > 0:
                eqws = eqconstraints.output_workspace()
                if not all(isinstance(w, OverloadedType) for w in eqws):
                    raise TypeError("""To use constraints with ROL the constraint value needs
    to be an OverloadedType.""")
                eqres = [ROLConstraint(eqconstraints)], [ROLVector(eqws)]
            else:
                eqres = [], []

            ineqconstraints = self.problem.constraints.inequality_constraints()
            if len(ineqconstraints.constraints) > 0:
                ineqws = ineqconstraints.output_workspace()
                if not all(isinstance(w, OverloadedType) for w in ineqws):
                    raise TypeError("""To use constraints with ROL the constraint value needs
    to be an OverloadedType.""")
                ineqres = [ROLConstraint(ineqconstraints)], [ROLVector(ineqws)]
            else:
                ineqres = [], []

            return eqres, ineqres

        @no_annotations
        def solve(self):
            """
            Solve the optimization problem and return the optimized
            parameters.
            """

            bnd = self.bounds
            econs = self.constraints[0][0]
            emuls = self.constraints[0][1]
            icons = self.constraints[1][0]
            imuls = self.constraints[1][1]
            if len(icons) > 0:
                zeros = [i.clone() for i in imuls]
                ibnds = [ROL.Bounds(z, isLower=True) for z in zeros]
            else:
                ibnds = []

            rolproblem = ROL.OptimizationProblem(self.rolobjective,
                                                 self.rolvector,
                                                 bnd=bnd,
                                                 econs=econs,
                                                 emuls=emuls,
                                                 icons=icons,
                                                 imuls=imuls,
                                                 ibnds=ibnds)
            x = self.rolvector
            params = ROL.ParameterList(self.params_dict, "Parameters")
            self.solver = ROL.OptimizationSolver(rolproblem, params)
            self.solver.solve()
            return self.problem.reduced_functional.controls.delist(x.dat)

        def checkGradient(self):
            x = self.rolvector
            g = x.clone()
            self.rolobjective.update(x, None, None)
            self.rolobjective.gradient(g, x, 0.0)
            res = self.rolobjective.checkGradient(x, g, 7, 1)
            return res

        def getAlgorithmState(self):
            return self.solver.getAlgorithmState()

except ImportError:

    class ROLSolver(object):
        def __init__(self, *args, **kwargs):
            raise ImportError("Could not import pyrol. Please install roltrilinos ROL using pip.")



from collections.abc import Hashable, Iterable, Iterator, MutableSet, Set


class OrderedSet(MutableSet):
    def __init__(self, iterable: Iterable[Hashable] = ()):
        """
        An OrderedSet is a set that maintains the insertion order
        of its elements.

        Args:
          iterable: An iterable with which to initialise the set elements.
        """

        self._elements: dict[Hashable, None] = {}

        for element in iterable:
            self._elements[element] = None

    def __contains__(self, obj: Hashable) -> bool:
        return obj in self._elements

    def __len__(self) -> int:
        return len(self._elements)

    def copy(self) -> "OrderedSet":
        """Return a shallow copy of this set"""

        ret = type(self)()
        ret._elements = self._elements.copy()

        return ret

    def add(self, obj: Hashable) -> None:
        """Add obj to this set."""

        self._elements[obj] = None

    def remove(self, obj: Hashable) -> None:
        """Remove obj from this set, it must be a member."""

        if obj not in self:
            raise KeyError(obj)

        del self._elements[obj]

    def discard(self, obj: Hashable) -> None:
        """Remove obj from this set if it is present."""

        if obj not in self:
            return

        self.remove(obj)

    def __iter__(self) -> Iterator:
        return iter(self._elements.keys())

    def union(self, *others: Iterable[Hashable]) -> "OrderedSet":
        """Return a new set with elements from this set and others."""

        ret = self.copy()

        for other in others:
            for element in other:
                ret.add(element)

        return ret

    def __or__(self, other: Set) -> "OrderedSet":
        if not isinstance(other, Set):
            raise TypeError

        return self.union(other)

    def difference(self, *others: Iterable[Hashable]) -> "OrderedSet":
        """Return a new set with elements from this set
        that are not in others.
        """

        ret = self.copy()

        for other in others:
            for element in other:
                ret.discard(element)

        return ret

    def __sub__(self, other: Set) -> "OrderedSet":
        if not isinstance(other, Set):
            raise TypeError

        return self.difference(other)

    def intersection(self, *others: Iterable[Hashable]) -> "OrderedSet":
        """Return a new set with elements common to this set
        and all others.
        """

        ret = type(self)()

        for element in self:
            for other in others:
                if element not in other:
                    break
            else:
                ret.add(element)

        return ret

    def __and__(self, other: Set) -> "OrderedSet":
        if not isinstance(other, Set):
            raise TypeError

        return self.intersection(other)

    def symmetric_difference(self, other: Iterable[Hashable]) -> "OrderedSet":
        """Return a new set with elements either in this set or other,
        but not both.
        """

        ret = type(self)()

        for element in self:
            if element not in other:
                ret.add(element)

        for element in other:
            if element not in self:
                ret.add(element)

        return ret

    def __xor__(self, other: Set) -> "OrderedSet":
        if not isinstance(other, Set):
            raise TypeError

        return self.symmetric_difference(other)
